"""SQL implementation"""

import sys
from collections.abc import Mapping, MutableMapping
from typing import Any, Dict, Iterable, Literal, TypeVar, Union

from pydantic import create_model
from pydantic.main import ModelT

from ._base import BaseStore
from ._compat import (
    AsyncSession,
    Column,
    DetachedInstanceError,
    IncEx,
    InstrumentedAttribute,
    RelationshipProperty,
    Table,
    _ColumnExpressionArgument,
    _ColumnExpressionOrStrLabelArgument,
    _SQLModel,
    create_async_engine,
    delete,
    insert,
    select,
    subqueryload,
    update,
)
from ._field import Field, get_field_definitions
from .query.parsers import QueryParser
from .query.selectors import QuerySelector

_T = TypeVar("_T", bound=_SQLModel)
_Filter = _ColumnExpressionArgument[bool] | bool


class SQLStore(BaseStore):
    """The store based on SQL relational database"""

    def __init__(self, uri: str, parser: QueryParser | None = None, **kwargs):
        super().__init__(uri, parser=parser, **kwargs)
        self._engine = create_async_engine(uri, **kwargs)

    async def register(self, models: list[type[_T]], checkfirst: bool = True):
        tables = [v.__table__ for v in models]
        async with self._engine.begin() as conn:
            await conn.run_sync(
                _SQLModel.metadata.create_all, tables=tables, checkfirst=checkfirst
            )

    async def insert(
        self, model: type[_T], items: Iterable[_T | dict], **kwargs
    ) -> list[_T]:
        parsed_items = [
            v if isinstance(v, model) else model.model_validate(v) for v in items
        ]
        relations_mapper = _get_relations_mapper(model)

        async with AsyncSession(self._engine) as session:
            stmt = insert(model).returning(model)
            cursor = await session.stream_scalars(stmt, parsed_items)
            results = await cursor.all()
            result_ids = [v.id for v in results]

            # insert embedded items also to permit something like
            # store.insert(Lib, [{"books": [{"title": "yay"}, ...]}])
            # where "books" is a one-to-many relationship
            # i.e. the kind that might be 'embedded' in Mongo-terms
            for k, field in relations_mapper.items():
                embedded_values = []

                for idx, record in enumerate(items):
                    parent = results[idx]
                    raw_value = _get_key_or_prop(record, k)
                    embedded_value = _parse_embedded(raw_value, field, parent)
                    if isinstance(embedded_value, Iterable):
                        embedded_values += embedded_value
                    elif isinstance(embedded_value, _SQLModel):
                        embedded_values.append(embedded_value)

                # insert the related items
                if len(embedded_values) > 0:
                    field_model = field.property.mapper.class_
                    embed_stmt = insert(field_model).returning(field_model)
                    await session.stream_scalars(embed_stmt, embedded_values)

            await session.commit()
            refreshed_results = await self.find(model, model.id.in_(result_ids))
            return list(refreshed_results)

    async def find(
        self,
        model: type[_T],
        *filters: _Filter,
        query: QuerySelector | None = None,
        skip: int = 0,
        limit: int | None = None,
        sort: tuple[_ColumnExpressionOrStrLabelArgument[Any]] = (),
        **kwargs,
    ) -> list[_T]:
        async with AsyncSession(self._engine) as session:
            if query:
                filters = (*filters, *self._parser.to_sql(model, query=query))

            relations = _get_relations(model)

            # eagerly load all relationships so that no validation errors occur due
            # to missing session if there is an attempt to load them lazily later
            eager_load_opts = [subqueryload(v) for v in relations]

            filtered_relations = _get_filtered_relations(
                filters=filters,
                relations=relations,
            )

            # Note that we need to treat relations that are referenced in the filters
            # differently from those that are not. This is because filtering basing on a relationship
            # requires the use of an inner join. Yet an inner join automatically excludes rows
            # that are have null for a given relationship.
            #
            # An outer join on the other hand would just return all the rows in the left table.
            # We thus need to do an inner join on tables that are being filtered.
            stmt = select(model)
            for rel in filtered_relations:
                stmt = stmt.join_from(model, rel)

            cursor = await session.stream_scalars(
                stmt.options(*eager_load_opts)
                .where(*filters)
                .limit(limit)
                .offset(skip)
                .order_by(*sort)
            )
            results = await cursor.all()
            return list(results)

    async def update(
        self,
        model: type[_T],
        *filters: _Filter,
        query: QuerySelector | None = None,
        updates: dict | None = None,
        **kwargs,
    ) -> list[_T]:
        async with AsyncSession(self._engine) as session:
            if query:
                filters = (*filters, *self._parser.to_sql(model, query=query))

            # Construct filters that have sub queries
            relations = _get_relations(model)
            rel_filters, non_rel_filters = _sieve_rel_from_non_rel_filters(
                filters=filters,
                relations=relations,
            )
            rel_filters = _to_subquery_based_filters(
                model=model,
                rel_filters=rel_filters,
                relations=relations,
            )

            # dealing with nested models in the update
            relations_mapper = _get_relations_mapper(model)
            embedded_updates = {}
            for k in relations_mapper:
                try:
                    embedded_updates[k] = updates.pop(k)
                except KeyError:
                    pass

            stmt = (
                update(model)
                .where(*non_rel_filters, *rel_filters)
                .values(**updates)
                .returning(model.__table__)
            )

            cursor = await session.stream(stmt)
            raw_results = await cursor.fetchall()
            results = [model.model_validate(row._mapping) for row in raw_results]
            result_ids = [v.id for v in results]

            for k, v in embedded_updates.items():
                field = relations_mapper[k]
                field_props = field.property
                field_model = field_props.mapper.class_
                # fk = foreign key
                fk_field_name = field_props.primaryjoin.right.name
                fk_field = getattr(field_model, fk_field_name)
                parent_id_field = field_props.primaryjoin.left.name

                # get the foreign keys to use in resetting all affected
                # relationships;
                # get parsed embedded values so that they can replace
                # the old relations.
                # Note: this operation is strictly replace, not patch
                embedded_values = []
                fk_values = []
                for parent in results:
                    embedded_value = _parse_embedded(v, field, parent)
                    if isinstance(embedded_value, Iterable):
                        embedded_values += embedded_value
                        fk_values.append(getattr(parent, parent_id_field))
                    elif isinstance(embedded_value, _SQLModel):
                        embedded_values.append(embedded_value)
                        fk_values.append(getattr(parent, parent_id_field))

                # insert the related items
                if len(embedded_values) > 0:
                    # Reset the relationship; delete all other related items
                    # Currently, this operation replaces all past relations
                    reset_stmt = delete(field_model).where(fk_field.in_(fk_values))
                    await session.stream(reset_stmt)

                    # insert the latest changes
                    embed_stmt = insert(field_model).returning(field_model)
                    await session.stream_scalars(embed_stmt, embedded_values)

            await session.commit()
            return await self.find(model, model.id.in_(result_ids))

    async def delete(
        self,
        model: type[_T],
        *filters: _Filter,
        query: QuerySelector | None = None,
        **kwargs,
    ) -> list[_T]:
        async with AsyncSession(self._engine) as session:
            if query:
                filters = (*filters, *self._parser.to_sql(model, query=query))

            deleted_items = await self.find(model, *filters)

            # Construct filters that have sub queries
            relations = _get_relations(model)
            rel_filters, non_rel_filters = _sieve_rel_from_non_rel_filters(
                filters=filters,
                relations=relations,
            )
            rel_filters = _to_subquery_based_filters(
                model=model,
                rel_filters=rel_filters,
                relations=relations,
            )
            exec_options = {}
            if len(rel_filters) > 0:
                exec_options = {"is_delete_using": True}

            await session.stream(
                delete(model)
                .where(*non_rel_filters, *rel_filters)
                .execution_options(**exec_options),
            )
            await session.commit()
            return deleted_items


class _SQLModelMeta(_SQLModel):
    """The base class for all SQL models"""

    id: int | None = Field(default=None, primary_key=True)

    def model_dump(
        self,
        *,
        mode: Union[Literal["json", "python"], str] = "python",
        include: IncEx = None,
        exclude: IncEx = None,
        context: Union[Dict[str, Any], None] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        round_trip: bool = False,
        warnings: Union[bool, Literal["none", "warn", "error"]] = True,
        serialize_as_any: bool = False,
    ) -> Dict[str, Any]:
        data = super().model_dump(
            mode=mode,
            include=include,
            exclude=exclude,
            context=context,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings,
            serialize_as_any=serialize_as_any,
        )
        relations_mappers = _get_relations_mapper(self.__class__)
        for k, field in relations_mappers.items():
            if exclude is None or k not in exclude:
                try:
                    value = getattr(self, k, None)
                except DetachedInstanceError:
                    # ignore lazy loaded values
                    continue

                if value is not None or not exclude_none:
                    data[k] = _serialize_embedded(
                        value,
                        field=field,
                        mode=mode,
                        context=context,
                        by_alias=by_alias,
                        exclude_unset=exclude_unset,
                        exclude_defaults=exclude_defaults,
                        exclude_none=exclude_none,
                        round_trip=round_trip,
                        warnings=warnings,
                        serialize_as_any=serialize_as_any,
                    )

        return data


def SQLModel(
    name: str,
    schema: type[ModelT],
    /,
    relationships: dict[str, type[Any] | type[Union[Any]]] = None,
) -> type[_SQLModelMeta] | type[ModelT]:
    """Creates a new SQLModel for the given schema for redis

    A new model can be defined by::

        Model = SQLModel("Model", Schema)

    Args:
        name: the name of the model
        schema: the schema from which the model is to be made
        relationships: a map of <name>:annotation for all relationships

    Returns:
        a SQLModel model class with the given name
    """
    fields = get_field_definitions(schema, relationships=relationships, is_for_sql=True)

    return create_model(
        name,
        # module of the calling function
        __module__=sys._getframe(1).f_globals["__name__"],
        __doc__=schema.__doc__,
        __cls_kwargs__={"table": True},
        __base__=(_SQLModelMeta,),
        **fields,
    )


def _get_relations(model: type[_SQLModel]):
    """Gets all the relational fields of the given model

    Args:
        model: the SQL model to inspect

    Returns:
        list of Fields that have associated relationships
    """
    return [
        v
        for v in model.__mapper__.all_orm_descriptors.values()
        if isinstance(v.property, RelationshipProperty)
    ]


def _get_relations_mapper(model: type[_SQLModel]) -> dict[str, Any]:
    """Gets all the relational fields with their names of the given model

    Args:
        model: the SQL model to inspect

    Returns:
        dict of (name, Field) that have associated relationships
    """
    return {
        k: v
        for k, v in model.__mapper__.all_orm_descriptors.items()
        if isinstance(v.property, RelationshipProperty)
    }


def _get_filtered_tables(filters: Iterable[_Filter]) -> list[Table]:
    """Retrieves the tables that have been referenced in the filters

    Args:
        filters: the tuple of filters to inspect

    Returns:
        the list of Table instances referenced in the filters
    """
    return [
        getattr(v, "table")
        for filter_ in filters
        for v in filter_.get_children()
        if isinstance(v, Column)
    ]


def _get_filtered_relations(
    filters: Iterable[_Filter], relations: Iterable[InstrumentedAttribute[Any]]
) -> list[InstrumentedAttribute[Any]]:
    """Retrieves the relations that have been referenced in the filters

    Args:
        filters: the tuple of filters to inspect
        relations: all relations present on the model

    Returns:
        the list of relations referenced in the filters
    """
    filtered_tables = _get_filtered_tables(filters)
    return [rel for rel in relations if rel.property.target in filtered_tables]


def _sieve_rel_from_non_rel_filters(
    filters: Iterable[_Filter], relations: Iterable[InstrumentedAttribute[Any]]
) -> tuple[list[_Filter], list[_Filter]]:
    """Separates relational filters from non-relational ones

    Args:
        filters: the tuple of filters to inspect
        relations: all relations present on the model

    Returns:
        tuple(rel, non_rel) where rel = list of relational filters,
            and non_rel = non-relational filters
    """
    rel_targets = [v.property.target for v in relations]
    rel = []
    non_rel = []

    for filter_ in filters:
        operands = filter_.get_children()
        if any([getattr(v, "table", None) in rel_targets for v in operands]):
            rel.append(filter_)
        else:
            non_rel.append(filter_)

    return rel, non_rel


def _to_subquery_based_filters(
    model: type[_SQLModel],
    rel_filters: list[_Filter],
    relations: Iterable[InstrumentedAttribute[Any]],
) -> list[_Filter]:
    """Converts filters to those that use subqueries to connect to other models

    This is especially important for update() and delete() which do not
    don't have `.join()` methods on them.

    Args:
        model: the model for which the subquery-based filters are to be generated
        rel_filters: the filters that have relationships in them
        relations: the relationship fields on the model

    Returns:
        list of filters that use subqueries to access other tables/models
    """
    # This is based on:
    # https://docs.sqlalchemy.org/en/20/orm/queryguide/dml.html#update-delete-with-custom-where-criteria-for-joined-table-inheritance
    if len(rel_filters) == 0:
        return []

    filtered_relations = _get_filtered_relations(
        filters=rel_filters,
        relations=relations,
    )

    # create the subquery collecting ids of model, with inner join to related models
    subquery = select(model.id)
    for rel in filtered_relations:
        subquery = subquery.join_from(model, rel)

    # return a filter checking model id against the returned ids
    return [model.id.in_(subquery.where(*rel_filters))]


def _get_key_or_prop(obj: dict | Any, name: str) -> Any:
    """Gets value of a key or property from a given object or dict

    Args:
        obj: the object from which to get the value
        name: the name of the key or property

    Returns:
        the value corresponding to the given key or property
    """
    if isinstance(obj, Mapping):
        return obj.get(name, None)
    else:
        return getattr(obj, name, None)


def _with_value(obj: dict | Any, field: str, value: Any) -> Any:
    """Sets the value of a key or property of a given object or dict

    Note: this mutates the object in-place

    Args:
        obj: the object
        field: the name of the key or property
        value: the value to set

    Returns:
        the mutated object
    """
    if isinstance(obj, MutableMapping):
        obj[field] = value
    else:
        setattr(obj, field, value)

    return obj


def _parse_embedded(
    value: Iterable[dict | Any] | dict | Any, field: Any, parent: _SQLModel
) -> Iterable[_SQLModel] | _SQLModel | None:
    """Parses embedded items that can be a single item or many into SQLModels

    Args:
        value: the value to parse
        field: the field on which these embedded items are
        parent: the parent SQLModel to which this value is attached

    Returns:
        An iterable of SQLModel instances or a single SQLModel instance
        or None if value is None
    """
    if value is None:
        return None

    props = field.property  # type: RelationshipProperty[Any]
    wrapper_type = props.collection_class
    field_model = props.mapper.class_
    fk_field = props.primaryjoin.right.name
    parent_id_field = props.primaryjoin.left.name
    fk_value = getattr(parent, parent_id_field)

    if issubclass(wrapper_type, (list, tuple, set)):
        # add a foreign key values to link back to parent
        return wrapper_type(
            [
                field_model.model_validate(_with_value(v, fk_field, fk_value))
                for v in value
            ]
        )
    elif wrapper_type is None:
        # add a foreign key value to link back to parent
        linked_value = _with_value(value, fk_field, fk_value)
        return field_model.model_validate(linked_value)

    raise NotImplementedError(
        f"relationship of type annotation {wrapper_type} not supported yet"
    )


def _serialize_embedded(
    value: Iterable[_SQLModel] | _SQLModel, field: Any, **kwargs
) -> Iterable[dict] | dict | None:
    """Serializes embedded items that can be a single item or many into SQLModels

    Args:
        value: the value to serialize
        field: the field on which these embedded items are
        kwargs: extra key-word args to pass to model_dump()

    Returns:
        An iterable of dicts or a single dict
        or None if value is None
    """
    if value is None:
        return None

    props = field.property  # type: RelationshipProperty[Any]
    wrapper_type = props.collection_class

    if issubclass(wrapper_type, (list, tuple, set)):
        # add a foreign key values to link back to parent
        return wrapper_type([v.model_dump(**kwargs) for v in value])
    elif wrapper_type is None:
        return value.model_dump(**kwargs)

    raise NotImplementedError(
        f"relationship of type annotation {wrapper_type} not supported yet"
    )



"""Schemas for the application"""

from pydantic import BaseModel

from nqlstore import Field, Relationship


class TodoList(BaseModel):
    """A list of Todos"""

    name: str = Field(index=True, full_text_search=True)
    todos: list["Todo"] = Relationship(back_populates="parent", default=[])


class Todo(BaseModel):
    """A single todo Item"""

    title: str = Field(index=True)
    is_complete: str = Field(default="0")
    parent_id: int | None = Field(
        default=None,
        foreign_key="sqltodolist.id",
        disable_on_mongo=True,
        disable_on_redis=True,
    )
    parent: TodoList | None = Relationship(
        back_populates="todos",
        disable_on_mongo=True,
        disable_on_redis=True,
    )



import copy
import importlib
import json
import re
from os import path
from typing import Any, TypeVar

from pydantic import BaseModel

from nqlstore._base import BaseStore
from nqlstore._compat import PydanticObjectId, _ColumnExpressionArgument
from nqlstore._sql import SQLModel, select

_SQLFilter = _ColumnExpressionArgument[bool] | bool
_TESTS_FOLDER = path.dirname(path.abspath(__file__))
_FIXTURES_PATH = path.join(_TESTS_FOLDER, "fixtures")
_LibType = TypeVar("_LibType", bound=BaseModel)
_BookType = TypeVar("_BookType", bound=BaseModel)


def load_fixture(fixture_name: str) -> list[dict[str, Any]] | dict[str, Any]:
    """Load fixture and return it as python objects

    Args:
        fixture_name: the name of the fixture file name

    Returns:
        the fixture as python objects
    """
    file_path = path.join(_FIXTURES_PATH, fixture_name)
    with open(file_path, "rb") as file:
        return json.load(file)


async def insert_test_data(
    store: BaseStore,
    library_model: type[_LibType],
    book_model: type[_BookType],
) -> list[_LibType]:
    """Insert data in the database before tests

    Args:
        store: the store to insert test data in
        library_model: the model class for the Library
        book_model: the model class for the Book

    Returns:
        the inserted libraries (with books embedded)
    """
    library_data = load_fixture("libraries.json")
    book_data = load_fixture("books.json")
    await store.register([library_model, book_model])

    library_data = _embed_test_books(book_model, libs=library_data, books=book_data)
    libraries = await store.insert(library_model, library_data)
    return libraries


def _populate_libs_with_books(
    libs: list[_LibType], books: list[_BookType]
) -> list[_LibType]:
    """Gets the list of libraries with their book lists updated to respective books

    Args:
        libs: the list of libraries
        books: the list of books

    Returns:
        list of libraries with their "books" property populated
    """
    books_per_lib = {}
    for bk in books:
        lib_books = books_per_lib.setdefault(bk.library_id, [])
        lib_books.append(bk)

    return [
        lib.model_copy(update={"books": books_per_lib.get(lib.id, [])}) for lib in libs
    ]


def _embed_test_books(
    model: type[_BookType], libs: list[dict], books: list[dict]
) -> list[dict]:
    """Embeds the books into the libraries

    Args:
        model: the model for the books
        libs: the list of library dicts
        books: the list of book dicts

    Returns:
        the list of library dicts with the books embedded
    """
    libs_copy = copy.deepcopy(libs)

    for idx, data in enumerate(books):
        try:
            libs_copy[idx % 2]["books"].append(model(**data))
        except KeyError:
            libs_copy[idx % 2]["books"] = [{**data}]

    return libs_copy


def _attach_test_books(
    model: type[_BookType], books: list[dict], libs: list[_LibType]
) -> list[_BookType]:
    """Attaches test books to libraries in a predetermined format

    Args:
        model: the model for the books
        books: the list of records to insert
        libs: the list of library instances to attach the books to

    Returns:
        the attached books
    """
    return [
        model(library_id=libs[idx % 2].id, **data) for idx, data in enumerate(books)
    ]


def to_sql_text(model: type[SQLModel], queries: tuple[_SQLFilter, ...]) -> str:
    """Converts a tuple of sql filters into their sql text

    It assumes the queries are to be used in a find operation

    Args:
        model: the model to operate on
        queries: the filters to use on the data

    Returns:
        the sql as text if this were for a find operation
    """
    sql = select(model).where(*queries)
    return str(sql.compile())


def get_regex_test_params(libs: list[_LibType]) -> list[tuple[dict, list[_LibType]]]:
    """Generates the test params for the REGEX test for the given libs

    Args:
        libs: the Library records

    Returns:
        pairs of regex filter and expected output after querying
    """
    return [
        (
            {"name": {"$regex": "^bu.*", "$options": "i"}},
            [v for v in libs if re.match(r"^bu.*", v.name, re.I)],
        ),
        (
            {"name": {"$regex": "^bu.*"}},
            [v for v in libs if re.match(r"^bu.*", v.name)],
        ),
        (
            {"name": {"$regex": "am.*"}},
            [v for v in libs if re.match(r".*am.*", v.name)],
        ),
        (
            {"name": {"$regex": ".*i$"}},
            [v for v in libs if re.match(r".*i$", v.name)],
        ),
    ]


def is_lib_installed(lib: str) -> bool:
    """Check if a library is installed.

    Args:
        lib: the library to check

    Returns:
        True if library is installed else False
    """
    return importlib.util.find_spec(lib) is not None



"""Module containing all possible query selectors

See: https://www.mongodb.com/docs/manual/reference/operator/query/#std-label-query-selectors
"""

from typing import Any, TypedDict

# Query Selectors

## Comparison
_EqSelector = TypedDict("_EqSelector", {"$eq": Any})
"""field is equal to value: ``{ $eq: <value> }``"""

_GtSelector = TypedDict("_GtSelector", {"$gt": Any})
"""field is greater than value: ``{ $gt: value }``"""

_GteSelector = TypedDict("_GteSelector", {"$gte": Any})
"""field is greater or equal to value: ``{ $gte: value }``"""

_InSelector = TypedDict("_InSelector", {"$in": list[Any]})
"""field is in list: ``{ $in: [<value1>, <value2>, ... <valueN> ] }``"""

_LtSelector = TypedDict("_LtSelector", {"$lt": Any})
"""field is less than value: ``{ $lt: value }``"""

_LteSelector = TypedDict("_LteSelector", {"$lte": Any})
"""field is less or equal to value: ``{ $lte: value }``"""

_NeSelector = TypedDict("_NeSelector", {"$ne": Any})
"""field is not equal to value: ``{ $ne: value }``"""

_NinSelector = TypedDict("_NinSelector", {"$nin": list[Any]})
"""field is not in list: ``{ $nin: [ <value1>, <value2> ... <valueN> ] }``"""


_RegexSelector = TypedDict(
    "_RegexSelector", {"$regex": str, "$options": str}, total=False
)
"""field matches given regular expression: ``{ "$regex": "pattern", "$options": "<options>" }``

``$options`` is optional
"""

## Element
_ExistsSelector = TypedDict("_ExistsSelector", {"$exists": bool})
"""field exists: ``{ $exists: <boolean> }``"""

_TypeSelector = TypedDict("_TypeSelector", {"$type": Any})
"""field is given BSON type: ``{ $type: <value1> }``"""

## Logical
OperatorSelector = (
    _EqSelector
    | _GtSelector
    | _GteSelector
    | _InSelector
    | _LtSelector
    | _LteSelector
    | _NeSelector
    | _NinSelector
    | _ExistsSelector
    | _TypeSelector
)
FieldSelector = dict[str, OperatorSelector]
"""a comparison on a given field: ``{ field: { <operator>: <operand>}}``"""

_AndSelector = TypedDict("_AndSelector", {"$and": list[FieldSelector]})
"""all expressions are true: ``{ $and: [ { <expression1> }, { <expression2> } , ... , { <expressionN> } ] }``"""

_NotSelector = TypedDict("_NotSelector", {"$not": OperatorSelector})
"""not true: ``{ field: { $not: { <operator-expression> } } }``"""

_NorSelector = TypedDict("_NorSelector", {"$nor": list[FieldSelector]})
"""no expression is true: ``{ $nor: [ { <expression1> }, { <expression2> }, ...  { <expressionN> } ] }``"""

_OrSelector = TypedDict("_OrSelector", {"$or": list[FieldSelector]})
"""any expression is true: ``{ $or: [ { <expression1> }, { <expression2> }, ... , { <expressionN> } ] }``"""

# FIXME: Add more starting at https://www.mongodb.com/docs/manual/reference/operator/query/#element

QuerySelector = FieldSelector | _AndSelector | _NotSelector | _NorSelector | _OrSelector
"""query predicate for filtering records"""



"""Module containing the registry of stores at runtime"""

import os

from models import (
    MongoTodo,
    MongoTodoList,
    RedisTodo,
    RedisTodoList,
    SqlTodo,
    SqlTodoList,
)

from nqlstore import MongoStore, RedisStore, SQLStore

_STORES: dict[str, MongoStore | RedisStore | SQLStore] = {}


async def get_redis_store() -> RedisStore | None:
    """Gets the redis store whose URL is specified in the environment"""
    global _STORES

    if redis_url := os.environ.get("REDIS_URL"):
        try:
            return _STORES[redis_url]
        except KeyError:
            store = RedisStore(uri=redis_url)
            await store.register([RedisTodoList, RedisTodo])
            _STORES[redis_url] = store
            return store


async def get_sql_store() -> SQLStore | None:
    """Gets the sql store whose URL is specified in the environment"""
    global _STORES

    if sql_url := os.environ.get("SQL_URL"):
        try:
            return _STORES[sql_url]
        except KeyError:
            store = SQLStore(uri=sql_url)
            await store.register([SqlTodoList, SqlTodo])
            _STORES[sql_url] = store
            return store


async def get_mongo_store() -> MongoStore | None:
    """Gets the mongo store whose URL and database are specified in the environment"""
    global _STORES

    if mongo_url := os.environ.get("MONGO_URL"):
        mongo_db = os.environ.get("MONGO_DB", "todos")
        mongo_full_url = f"{mongo_url}/{mongo_db}"

        try:
            return _STORES[mongo_full_url]
        except KeyError:
            store = MongoStore(uri=mongo_url, database=mongo_db)
            await store.register([MongoTodoList, MongoTodo])
            _STORES[mongo_full_url] = store
            return store


def clear_stores():
    """Clears the registry of stores

    Important for clean up
    """
    global _STORES
    _STORES.clear()



import pytest
import pytest_asyncio
from pydantic import BaseModel

from nqlstore import (
    EmbeddedJsonModel,
    EmbeddedMongoModel,
    Field,
    JsonModel,
    MongoModel,
    MongoStore,
    RedisStore,
    Relationship,
    SQLModel,
    SQLStore,
)
from nqlstore.query.parsers import QueryParser

from .utils import get_regex_test_params, insert_test_data, is_lib_installed


class Library(BaseModel):
    address: str = Field(index=True, full_text_search=True)
    name: str = Field(index=True, full_text_search=True)
    books: list["Book"] = Relationship(back_populates="library", default=[])

    class Settings:
        name = "libraries"


class Book(BaseModel):
    title: str = Field(index=True)
    library_id: int | None = Field(
        default=None,
        foreign_key="sqllibrary.id",
        disable_on_redis=True,
        disable_on_mongo=True,
    )
    library: Library | None = Relationship(
        back_populates="books", disable_on_redis=True, disable_on_mongo=True
    )


# default models
SqlLibrary = Library
SqlBook = Book
MongoLibrary = Library
MongoBook = Book
RedisLibrary = Library
RedisBook = Book

if is_lib_installed("sqlmodel"):
    SqlLibrary = SQLModel(
        "SqlLibrary", Library, relationships={"books": list["SqlBook"]}
    )
    SqlBook = SQLModel("SqlBook", Book, relationships={"library": SqlLibrary | None})

if is_lib_installed("beanie"):
    MongoBook = EmbeddedMongoModel("MongoBook", Book)
    MongoLibrary = MongoModel(
        "MongoLibrary", Library, embedded_models={"books": list[MongoBook]}
    )

if is_lib_installed("redis_om"):
    RedisBook = EmbeddedJsonModel("RedisBook", Book)
    RedisLibrary = JsonModel(
        "RedisLibrary", Library, embedded_models={"books": list[RedisBook]}
    )


@pytest.fixture
@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
def sql_store():
    """The sql store stored in memory"""
    store = SQLStore(uri="sqlite+aiosqlite:///:memory:")
    yield store


@pytest.fixture
@pytest.mark.skipif(not is_lib_installed("beanie"), reason="Requires beanie.")
def mongo_store():
    """The mongodb store"""
    import pymongo

    store = MongoStore(uri="mongodb://localhost:27017", database="testing")
    yield store

    # clean up after the test
    client = pymongo.MongoClient("mongodb://localhost:27017")
    client.drop_database("testing")


@pytest.fixture
@pytest.mark.skipif(not is_lib_installed("redis_om"), reason="Requires redis_om.")
def redis_store():
    """The redis store"""
    import redis

    store = RedisStore(uri="redis://localhost:6379/0")
    yield store

    # clean up after the test
    client = redis.Redis("localhost", 6379, 0)
    client.flushall()


@pytest_asyncio.fixture()
@pytest.mark.skipif(not is_lib_installed("redis_om"), reason="Requires redis_om.")
async def inserted_redis_libs(redis_store):
    """The libraries inserted in the redis store"""
    inserted_libs = await insert_test_data(
        redis_store,
        library_model=RedisLibrary,
        book_model=RedisBook,
    )
    # sanity check
    all_books = [bk for lib in inserted_libs for bk in lib.books]
    assert all_books != []

    yield inserted_libs


@pytest.fixture
@pytest.mark.skipif(not is_lib_installed("redis_om"), reason="Requires redis_om.")
def regex_params_redis(inserted_redis_libs):
    """The regex test params for redis"""
    yield get_regex_test_params(inserted_redis_libs)


@pytest_asyncio.fixture()
@pytest.mark.skipif(not is_lib_installed("beanie"), reason="Requires beanie.")
async def inserted_mongo_libs(mongo_store):
    """The libraries inserted in the mongodb store"""
    inserted_libs = await insert_test_data(
        mongo_store, library_model=MongoLibrary, book_model=MongoBook
    )
    # sanity check
    all_books = [bk for lib in inserted_libs for bk in lib.books]
    assert all_books != []

    yield inserted_libs


@pytest.fixture
@pytest.mark.skipif(not is_lib_installed("beanie"), reason="Requires beanie.")
def regex_params_mongo(inserted_mongo_libs):
    """The regex test params for mongo"""
    yield get_regex_test_params(inserted_mongo_libs)


@pytest_asyncio.fixture()
@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
async def inserted_sql_libs(sql_store):
    """The libraries inserted in the sql store"""
    inserted_libs = await insert_test_data(
        sql_store, library_model=SqlLibrary, book_model=SqlBook
    )
    # sanity check
    all_books = [bk for lib in inserted_libs for bk in lib.books]
    assert all_books != []

    yield inserted_libs


@pytest.fixture
@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
def regex_params_sql(inserted_sql_libs):
    """The regex test params for sql"""
    yield get_regex_test_params(inserted_sql_libs)


@pytest_asyncio.fixture()
@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
async def sql_qparser():
    """The default query parser for sql"""
    qparser = QueryParser()
    sql_store = SQLStore(uri="sqlite+aiosqlite:///:memory:", parser=qparser)
    await sql_store.register([SqlLibrary, SqlBook])
    yield qparser


@pytest_asyncio.fixture()
@pytest.mark.skipif(not is_lib_installed("redis_om"), reason="Requires redis_om.")
async def redis_qparser():
    """The default query parser for redis"""
    import redis

    qparser = QueryParser()
    redis_store = RedisStore(uri="redis://localhost:6379/0", parser=qparser)
    await redis_store.register([RedisLibrary, RedisBook])

    yield qparser

    # clean up after the test
    client = redis.Redis("localhost", 6379, 0)
    client.flushall()



"""The module with the abstract classes for this package"""

import abc
from typing import Any, Iterable, TypeVar

from pydantic import BaseModel

from nqlstore.query.parsers import QueryParser
from nqlstore.query.selectors import QuerySelector

_T = TypeVar("_T", bound=BaseModel)


class BaseStore(abc.ABC):
    """Abstract class for storing data"""

    def __init__(self, uri: str, parser: QueryParser | None = None, **kwargs):
        """
        Args:
            uri: the URI to the underlying store
            parser: the query parser for parsing NQL mongodb-like queries.
            kwargs: extra key-word args to pass to the initializer
        """
        if parser is None:
            parser = QueryParser()

        self._parser = parser
        self._uri = uri

    @abc.abstractmethod
    async def register(self, models: list[type[_T]], **kwargs):
        """Registers the given models and runs any initialization steps

        Args:
            models: the list of Model's this store is to contain
            kwargs: extra key-word args to pass to the initializer
        """

    @abc.abstractmethod
    async def insert(
        self, model: type[_T], items: Iterable[_T | dict], **kwargs
    ) -> list[_T]:
        """Inserts the items to the store

        Args:
            model: the model whose instances are being inserted
            items: the items to insert into the store
            kwargs: extra key-word arguments

        Returns:
            the created items
        """

    @abc.abstractmethod
    async def find(
        self,
        model: type[_T],
        *filters: Any,
        query: QuerySelector | None = None,
        skip: int = 0,
        limit: int | None = None,
        sort: Any = None,
        **kwargs,
    ) -> list[_T]:
        """Find the items that fulfill the given filters

        Args:
            filters: the things to match against
            query: alternative mongodb-like query object to us alongside or instead of native filters
            model: the model whose instances are being queried
            skip: number of records to ignore at the top of the returned results; default is 0
            limit: maximum number of records to return; default is None.
            sort: fields to sort by; default = None
            kwargs: extra key-word args to pass to the underlying find method

        Returns:
            the matched items
        """

    @abc.abstractmethod
    async def update(
        self,
        model: type[_T],
        *filters: Any,
        query: QuerySelector | None = None,
        updates: dict | None = None,
        **kwargs,
    ) -> list[_T]:
        """Update the items that fulfill the given filters

        Args:
            filters: the things to match against
            query: alternative mongodb-like query object to us alongside or instead of native filters
            updates: the payload to update the items with
            model: the model whose instances are being updated
            kwargs: extra key-word args to pass to the underlying update method

        Returns:
            the items after updating
        """

    @abc.abstractmethod
    async def delete(
        self,
        model: type[_T],
        *filters: Any,
        query: QuerySelector | None = None,
        **kwargs,
    ) -> list[_T]:
        """Delete the items that fulfill the given filters

        Args:
            filters: the things to match against
            query: alternative mongodb-like query object to us alongside or instead of native filters
            model: the model whose instances are being deleted
            kwargs: extra key-word args to pass to the underlying delete method

        Returns:
            the deleted items
        """



import pytest

from tests.conftest import SqlBook, SqlLibrary
from tests.utils import is_lib_installed, load_fixture

_LIBRARY_DATA = load_fixture("libraries.json")
_TEST_ADDRESS = "Hoima, Uganda"


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
async def test_model_dump(sql_store, inserted_sql_libs):
    """model_dump should recursively dump any 'embedded' models"""
    got = [v.model_dump() for v in inserted_sql_libs]
    expected = [
        {
            **v.model_dump(exclude={"books"}),
            "books": [bk.model_dump() for bk in v.books],
        }
        for v in inserted_sql_libs
    ]
    assert got == expected


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
async def test_find_native(sql_store, inserted_sql_libs):
    """Find should return the items that match the native filter"""
    got = await sql_store.find(
        SqlLibrary,
        (SqlLibrary.address == _TEST_ADDRESS) | (SqlLibrary.name.startswith("Ba")),
        skip=1,
    )
    expected = [
        v
        for v in inserted_sql_libs
        if v.address == _TEST_ADDRESS or v.name.startswith("Ba")
    ][1:]
    assert _ordered(got) == _ordered(expected)


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
async def test_find_mongo_style(sql_store, inserted_sql_libs):
    """Find should return the items that match the mongodb-like filter"""
    got = await sql_store.find(
        SqlLibrary,
        query={"$or": [{"address": {"$eq": _TEST_ADDRESS}}, {"name": {"$eq": "Bar"}}]},
        skip=1,
    )
    expected = [
        v for v in inserted_sql_libs if v.address == _TEST_ADDRESS or v.name == "Bar"
    ][1:]
    assert _ordered(got) == _ordered(expected)


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
async def test_find_dot_notation(sql_store, inserted_sql_libs):
    """Find should find the items that match the filter with embedded objects"""
    wanted_titles = ["Belljar", "Benediction man"]
    matches_query = lambda v: any(bk.title in wanted_titles for bk in v.books)

    got = await sql_store.find(
        SqlLibrary, query={"books.title": {"$in": wanted_titles}}
    )

    expected = [v for v in inserted_sql_libs if matches_query(v)]
    assert _ordered(got) == _ordered(expected)


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
@pytest.mark.parametrize("index", range(4))
async def test_regex_find_mongo_style(sql_store, regex_params_sql, index):
    """Find with regex should find the items that match the regex"""
    filters, expected = regex_params_sql[index]
    got = await sql_store.find(SqlLibrary, query=filters)
    assert _ordered(got) == _ordered(expected)


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
async def test_find_hybrid(sql_store, inserted_sql_libs):
    """Find should return the items that match the mongodb-like filter AND the native filter"""
    got = await sql_store.find(
        SqlLibrary,
        (SqlLibrary.name.startswith("Ba")),
        query={"address": {"$eq": _TEST_ADDRESS}},
        skip=1,
    )
    expected = [
        v
        for v in inserted_sql_libs
        if v.address == _TEST_ADDRESS and v.name.startswith("Ba")
    ][1:]
    assert _ordered(got) == _ordered(expected)


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
async def test_create(sql_store):
    """Create should add many items to the sql database"""
    await sql_store.register([SqlLibrary, SqlBook])
    got = await sql_store.insert(SqlLibrary, _LIBRARY_DATA)
    expected = [
        SqlLibrary(id=idx + 1, **item) for idx, item in enumerate(_LIBRARY_DATA)
    ]
    assert _ordered(got) == _ordered(expected)


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
async def test_update_native(sql_store, inserted_sql_libs):
    """Update should update the items that match the native filter"""
    updates = {
        "address": "some new address",
        "books": [
            {"title": "Upon this mountain"},
            {"title": "No longer at ease"},
        ],
    }
    matches_query = lambda v: v.name.startswith("Bu") and v.address == _TEST_ADDRESS

    # in immediate response
    # NOTE: redis startswith/contains on single letters is not supported by redis
    got = await sql_store.update(
        SqlLibrary,
        (SqlLibrary.name.startswith("Bu") & (SqlLibrary.address == _TEST_ADDRESS)),
        updates=updates,
    )
    expected = [
        record.model_copy(update=updates)
        for record in inserted_sql_libs
        if matches_query(record)
    ]
    assert _ordered(got) == _ordered(expected)

    # all library data in database
    got = await sql_store.find(SqlLibrary)
    expected = [
        (record.model_copy(update=updates) if matches_query(record) else record)
        for record in inserted_sql_libs
    ]
    assert _ordered(got) == _ordered(expected)


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
async def test_update_mongo_style(sql_store, inserted_sql_libs):
    """Update should update the items that match the mongodb-like filter"""
    updates = {
        "address": "some new address",
        "books": [
            {"title": "Upon this mountain"},
            {"title": "No longer at ease"},
        ],
    }
    matches_query = lambda v: v.name != "Kisaasi" and v.address == _TEST_ADDRESS

    # in immediate response
    # NOTE: redis startswith/contains on single letters is not supported by redis
    got = await sql_store.update(
        SqlLibrary,
        query={
            "$and": [
                {"name": {"$not": {"$eq": "Kisaasi"}}},
                {"address": {"$eq": _TEST_ADDRESS}},
            ]
        },
        updates=updates,
    )
    expected = [
        record.model_copy(update=updates)
        for record in inserted_sql_libs
        if matches_query(record)
    ]

    assert _ordered(got) == _ordered(expected)

    # all library data in database
    got = await sql_store.find(SqlLibrary)
    expected = [
        (record.model_copy(update=updates) if matches_query(record) else record)
        for record in inserted_sql_libs
    ]
    assert _ordered(got) == _ordered(expected)


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
async def test_update_hybrid(sql_store, inserted_sql_libs):
    """Update should update the items that match the mongodb-like filter AND the native filter"""
    updates = {
        "address": "some new address",
        "books": [
            {"title": "Upon this mountain"},
            {"title": "No longer at ease"},
        ],
    }
    matches_query = lambda v: v.name.startswith("Bu") and v.address == _TEST_ADDRESS

    # in immediate response
    # NOTE: redis startswith/contains on single letters is not supported by redis
    got = await sql_store.update(
        SqlLibrary,
        (SqlLibrary.name.startswith("Bu")),
        query={"address": {"$eq": _TEST_ADDRESS}},
        updates=updates,
    )
    expected = [
        record.model_copy(update=updates)
        for record in inserted_sql_libs
        if matches_query(record)
    ]
    assert _ordered(got) == _ordered(expected)

    # all library data in database
    got = await sql_store.find(SqlLibrary)
    expected = [
        (record.model_copy(update=updates) if matches_query(record) else record)
        for record in inserted_sql_libs
    ]
    assert _ordered(got) == _ordered(expected)


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
async def test_update_dot_notation(sql_store, inserted_sql_libs):
    """Update should update the items that match the filter with embedded objects"""
    wanted_titles = ["Belljar", "Benediction man"]
    updates = {
        "address": "some new address",
        "books": [
            {"title": "Upon this mountain"},
            {"title": "No longer at ease"},
        ],
    }
    matches_query = lambda v: any(bk.title in wanted_titles for bk in v.books)

    got = await sql_store.update(
        SqlLibrary,
        query={"books.title": {"$in": wanted_titles}},
        updates=updates,
    )
    expected = [
        record.model_copy(update=updates)
        for record in inserted_sql_libs
        if matches_query(record)
    ]
    assert _ordered(got) == _ordered(expected)

    # all library data in database
    got = await sql_store.find(SqlLibrary)
    expected = [
        (record.model_copy(update=updates) if matches_query(record) else record)
        for record in inserted_sql_libs
    ]
    assert _ordered(got) == _ordered(expected)


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
async def test_delete_native(sql_store, inserted_sql_libs):
    """Delete should delete the items that match the native filter"""
    # in immediate response
    # NOTE: redis startswith/contains on single letters is not supported by redis
    got = await sql_store.delete(SqlLibrary, SqlLibrary.address.startswith("Ho"))
    expected = [v for v in inserted_sql_libs if v.address.startswith("Ho")]
    assert _ordered(got) == _ordered(expected)

    # all data in database
    got = await sql_store.find(SqlLibrary)
    expected = [v for v in inserted_sql_libs if not v.address.startswith("Ho")]
    assert _ordered(got) == _ordered(expected)


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
async def test_delete_mongo_style(sql_store, inserted_sql_libs):
    """Delete should delete the items that match the mongodb-like filter"""
    addresses = ["Bujumbura, Burundi", "Non existent"]
    unwanted_names = ["Bar", "Kisaasi"]

    # in immediate response
    # NOTE: redis startswith/contains on single letters is not supported by redis
    got = await sql_store.delete(
        SqlLibrary,
        query={
            "$or": [
                {"$nor": [{"name": {"$eq": name}} for name in unwanted_names]},
                {"address": {"$in": addresses}},
            ]
        },
    )
    expected = [
        v
        for v in inserted_sql_libs
        if v.address in addresses or v.name not in unwanted_names
    ]
    assert _ordered(got) == _ordered(expected)

    # all data in database
    got = await sql_store.find(SqlLibrary)
    expected = [
        v
        for v in inserted_sql_libs
        if v.address not in addresses and v.name in unwanted_names
    ]
    assert _ordered(got) == _ordered(expected)


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
async def test_delete_hybrid(sql_store, inserted_sql_libs):
    """Delete should delete the items that match the mongodb-like filter AND the native filter"""
    unwanted_addresses = ["Stockholm, Sweden"]

    # in immediate response
    # NOTE: redis startswith/contains on single letters is not supported by redis
    got = await sql_store.delete(
        SqlLibrary,
        (SqlLibrary.name.startswith("bu")),
        query={"address": {"$nin": unwanted_addresses}},
    )
    expected = [
        v
        for v in inserted_sql_libs
        if v.address not in unwanted_addresses and v.name.lower().startswith("bu")
    ]
    assert _ordered(got) == _ordered(expected)

    # all data in database
    got = await sql_store.find(SqlLibrary)
    expected = [
        v
        for v in inserted_sql_libs
        if v.address in unwanted_addresses or not v.name.lower().startswith("bu")
    ]
    assert got == expected


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
async def test_delete_dot_notation(sql_store, inserted_sql_libs):
    """Delete should delete the items that match the filter with embedded objects"""
    wanted_titles = ["Belljar", "Benediction man"]
    matches_query = lambda v: any(bk.title in wanted_titles for bk in v.books)

    got = await sql_store.delete(
        SqlLibrary,
        query={"books.title": {"$in": wanted_titles}},
    )
    expected = [record for record in inserted_sql_libs if matches_query(record)]
    assert _ordered(got) == _ordered(expected)

    # all library data in database
    got = await sql_store.find(SqlLibrary)
    expected = [record for record in inserted_sql_libs if not matches_query(record)]
    assert _ordered(got) == _ordered(expected)


def _ordered(libs: list[SqlLibrary]) -> list[SqlLibrary]:
    """Sorts the libraries by id and returns them

    Args:
        libs: the library instances to sort

    Returns:
        the ordered libraries
    """
    return sorted(libs, key=lambda v: v.id)



import re

import pytest

from nqlstore._compat import Document
from tests.conftest import MongoBook, MongoLibrary
from tests.utils import is_lib_installed, load_fixture

_LIBRARY_DATA = load_fixture("libraries.json")
_BOOK_DATA = load_fixture("books.json")


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("beanie"), reason="Requires beanie.")
async def test_register(mongo_store, inserted_mongo_libs):
    """Register ensures that the MongoLibrary is properly initialized"""
    assert (
        MongoLibrary._document_settings.motor_collection.full_name
        == "testing.libraries"
    )
    assert not issubclass(MongoBook, Document)


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("beanie"), reason="Requires beanie.")
async def test_find(mongo_store, inserted_mongo_libs):
    """Find should find the items that match the filter"""
    got = await mongo_store.find(MongoLibrary, {}, skip=1)
    expected = [v for idx, v in enumerate(inserted_mongo_libs) if idx >= 1]
    assert got == expected


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("beanie"), reason="Requires beanie.")
async def test_find_dot_notation(mongo_store, inserted_mongo_libs):
    """Find should find the items that match the filter with embedded objects"""
    got = await mongo_store.find(
        MongoLibrary, {"books.title": {"$regex": "^be.*", "$options": "i"}}
    )
    expected = [
        v
        for v in inserted_mongo_libs
        if any([bk.title.lower().startswith("be") for bk in v.books])
    ]
    assert got == expected


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("beanie"), reason="Requires beanie.")
@pytest.mark.parametrize("index", range(4))
async def test_regex_find(mongo_store, regex_params_mongo, index):
    """Find with regex should find the items that match the regex"""
    filters, expected = regex_params_mongo[index]
    got = await mongo_store.find(MongoLibrary, filters)
    assert got == expected


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("beanie"), reason="Requires beanie.")
async def test_create(mongo_store):
    """Create should add many items to the mongo database"""
    await mongo_store.register([MongoLibrary])
    books = [MongoBook(**v) for v in _BOOK_DATA]
    lib_data = [{**v, "books": [*books]} for v in _LIBRARY_DATA]
    got = await mongo_store.insert(MongoLibrary, lib_data)
    got = [v.model_dump(exclude={"id"}) for v in got]
    expected = [{**v, "books": [*_BOOK_DATA]} for v in _LIBRARY_DATA]
    assert got == expected


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("beanie"), reason="Requires beanie.")
async def test_update(mongo_store, inserted_mongo_libs):
    """Update should update the items that match the filter"""
    updates = {"address": "some new address"}
    filters = {"name": re.compile(r"^b", re.I)}
    startswith_b = lambda v: v.name.lower().startswith("b")
    expected_data_in_db = [
        (record.model_copy(update=updates) if startswith_b(record) else record)
        for record in inserted_mongo_libs
    ]

    # in immediate response
    got = await mongo_store.update(MongoLibrary, filters, updates=updates)
    expected = list(filter(startswith_b, expected_data_in_db))
    assert got == expected

    # all library data in database
    got = await mongo_store.find(MongoLibrary, {})
    assert got == expected_data_in_db


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("beanie"), reason="Requires beanie.")
async def test_update_native(mongo_store, inserted_mongo_libs):
    """Update should update the items that match the filter with mongo-style update operators"""
    updates = {"$set": {"address": "some new address"}}
    filters = {"name": re.compile(r"^b", re.I)}
    startswith_b = lambda v: v.name.lower().startswith("b")
    expected_data_in_db = [
        (record.model_copy(update=updates["$set"]) if startswith_b(record) else record)
        for record in inserted_mongo_libs
    ]

    # in immediate response
    got = await mongo_store.update(MongoLibrary, filters, updates=updates)
    expected = list(filter(startswith_b, expected_data_in_db))
    assert got == expected

    # all library data in database
    got = await mongo_store.find(MongoLibrary, {})
    assert got == expected_data_in_db


@pytest.mark.asyncio
@pytest.mark.skipif(not is_lib_installed("beanie"), reason="Requires beanie.")
async def test_delete(mongo_store, inserted_mongo_libs):
    """Delete should remove the items that match the filter"""
    # in immediate response
    got = await mongo_store.delete(MongoLibrary, {"name": re.compile(r"^b", re.I)})
    expected = [v for v in inserted_mongo_libs if v.name.lower().startswith("b")]
    assert got == expected

    # all data in database
    got = await mongo_store.find(MongoLibrary, {})
    expected = [v for v in inserted_mongo_libs if not v.name.lower().startswith("b")]
    assert got == expected



"""Module to cater for missing dependencies"""

from typing import Any, Callable, Literal

"""
redis imports; and their defaults if redis_om is missing
"""
try:
    from aredis_om import EmbeddedJsonModel as _EmbeddedJsonModel
    from aredis_om import HashModel as _HashModel
    from aredis_om import JsonModel as _JsonModel
    from aredis_om import KNNExpression, Migrator
    from aredis_om import RedisModel as _RedisModel
    from aredis_om import get_redis_connection
    from aredis_om.model.model import Expression
    from aredis_om.model.model import Field as _RedisField
    from aredis_om.model.model import FieldInfo as _RedisFieldInfo
    from aredis_om.model.model import VectorFieldOptions, verify_pipeline_response
    from redis.asyncio import Redis
    from redis.client import Pipeline
except ImportError:
    from pydantic.fields import Field as _RedisField
    from pydantic.fields import FieldInfo as _RedisFieldInfo
    from pydantic.main import BaseModel

    VectorFieldOptions = Any
    _EmbeddedJsonModel = BaseModel
    _HashModel = BaseModel
    _JsonModel = BaseModel
    _RedisModel = BaseModel
    KNNExpression = Any
    Expression = Any
    Pipeline = Any
    Migrator = lambda *a, **k: dict(**k)
    get_redis_connection = lambda *a, **k: dict(**k)
    verify_pipeline_response = lambda *a, **k: dict(**k)
    Redis = Any


"""
sql imports; and their default if sqlmodel is missing
"""
try:
    from sqlalchemy import Column, Table
    from sqlalchemy.ext.asyncio import create_async_engine
    from sqlalchemy.orm import InstrumentedAttribute, RelationshipProperty, subqueryload
    from sqlalchemy.orm.exc import DetachedInstanceError
    from sqlalchemy.sql._typing import (
        _ColumnExpressionArgument,
        _ColumnExpressionOrStrLabelArgument,
    )
    from sqlmodel import SQLModel as _SQLModel
    from sqlmodel import delete, insert, select, update
    from sqlmodel._compat import post_init_field_info
    from sqlmodel.ext.asyncio.session import AsyncSession
    from sqlmodel.main import Field as _SQLField
    from sqlmodel.main import FieldInfo as _SqlFieldInfo
    from sqlmodel.main import IncEx, NoArgAnyCallable, OnDeleteType
    from sqlmodel.main import RelationshipInfo as _RelationshipInfo
except ImportError:
    from typing import Mapping, Optional, Sequence
    from typing import Set
    from typing import Set as _ColumnExpressionArgument
    from typing import Set as _ColumnExpressionOrStrLabelArgument
    from typing import Union

    from pydantic import BaseModel
    from pydantic._internal._repr import Representation
    from pydantic.fields import Field as _SQLField
    from pydantic.fields import FieldInfo as _FieldInfo

    _SQLModel = BaseModel
    post_init_field_info = lambda b: b
    NoArgAnyCallable = Callable[[], Any]
    OnDeleteType = Literal["CASCADE", "SET NULL", "RESTRICT"]
    Column = Any
    create_async_engine = lambda *a, **k: dict(**k)
    delete = insert = select = update = create_async_engine
    AsyncSession = Any
    RelationshipProperty = Set
    Table = Set
    InstrumentedAttribute = Set
    subqueryload = lambda *a, **kwargs: dict(**kwargs)
    DetachedInstanceError = RuntimeError
    IncEx = Set[Any] | dict

    class _SqlFieldInfo(_FieldInfo): ...

    class _RelationshipInfo(Representation):
        def __init__(
            self,
            *,
            back_populates: Optional[str] = None,
            cascade_delete: Optional[bool] = False,
            passive_deletes: Optional[Union[bool, Literal["all"]]] = False,
            link_model: Optional[Any] = None,
            sa_relationship: Optional[RelationshipProperty] = None,  # type: ignore
            sa_relationship_args: Optional[Sequence[Any]] = None,
            sa_relationship_kwargs: Optional[Mapping[str, Any]] = None,
        ) -> None:
            if sa_relationship is not None:
                if sa_relationship_args is not None:
                    raise RuntimeError(
                        "Passing sa_relationship_args is not supported when "
                        "also passing a sa_relationship"
                    )
                if sa_relationship_kwargs is not None:
                    raise RuntimeError(
                        "Passing sa_relationship_kwargs is not supported when "
                        "also passing a sa_relationship"
                    )
            self.back_populates = back_populates
            self.cascade_delete = cascade_delete
            self.passive_deletes = passive_deletes
            self.link_model = link_model
            self.sa_relationship = sa_relationship
            self.sa_relationship_args = sa_relationship_args
            self.sa_relationship_kwargs = sa_relationship_kwargs


"""
mongo imports; and their defaults if the 'beanie' package is not installed
"""
try:
    from beanie import Document, PydanticObjectId, SortDirection, init_beanie
    from motor.motor_asyncio import (
        AsyncIOMotorClient,
        AsyncIOMotorClientSession,
        AsyncIOMotorCollection,
    )
except ImportError:
    from pydantic import BaseModel

    init_beanie = lambda *a, **k: dict(**k)
    SortDirection = Any
    AsyncIOMotorClient = lambda *a, **k: dict(**k)
    AsyncIOMotorClientSession = Any
    AsyncIOMotorCollection = Any
    PydanticObjectId = Any
    Document = BaseModel



from ._field import Field, Relationship
from ._mongo import EmbeddedMongoModel, MongoModel, MongoStore, PydanticObjectId
from ._redis import EmbeddedJsonModel, HashModel, JsonModel, RedisStore
from ._sql import SQLModel, SQLStore
from .query.parsers import QueryParser

__all__ = [
    "SQLStore",
    "SQLModel",
    "MongoModel",
    "MongoStore",
    "PydanticObjectId",
    "EmbeddedMongoModel",
    "Field",
    "Relationship",
    "RedisStore",
    "HashModel",
    "JsonModel",
    "EmbeddedJsonModel",
    "QueryParser",
    "query",
]



"""The main module containing the default types to be imported"""

from copy import copy
from typing import (
    AbstractSet,
    Any,
    Dict,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Type,
    Union,
    overload,
)

from pydantic.main import ModelT
from pydantic_core import PydanticUndefined as Undefined
from pydantic_core import PydanticUndefinedType as UndefinedType

from ._compat import (
    Column,
    NoArgAnyCallable,
    OnDeleteType,
    RelationshipProperty,
    VectorFieldOptions,
    _RedisFieldInfo,
    _RelationshipInfo,
    _SqlFieldInfo,
    post_init_field_info,
)


class FieldInfo(_SqlFieldInfo, _RedisFieldInfo):
    def __init__(self, default: Any = Undefined, **kwargs: Any) -> None:
        disable_on_redis = kwargs.get("disable_on_redis", False)
        disable_on_sql = kwargs.get("disable_on_sql", False)
        disable_on_mongo = kwargs.get("disable_on_mongo", False)
        super().__init__(default=default, **kwargs)
        self.disable_on_redis = disable_on_redis
        self.disable_on_sql = disable_on_sql
        self.disable_on_mongo = disable_on_mongo


class RelationshipInfo(_RelationshipInfo):
    def __init__(
        self,
        *,
        back_populates: Optional[str] = None,
        cascade_delete: Optional[bool] = False,
        passive_deletes: Optional[Union[bool, Literal["all"]]] = False,
        link_model: Optional[Any] = None,
        sa_relationship: Optional[RelationshipProperty] = None,  # type: ignore
        sa_relationship_args: Optional[Sequence[Any]] = None,
        sa_relationship_kwargs: Optional[Mapping[str, Any]] = None,
        **kwargs: Any,
    ):
        disable_on_redis = kwargs.get("disable_on_redis", False)
        disable_on_sql = kwargs.get("disable_on_sql", False)
        disable_on_mongo = kwargs.get("disable_on_mongo", False)
        default = kwargs.get("default", Undefined)
        super().__init__(
            back_populates=back_populates,
            cascade_delete=cascade_delete,
            passive_deletes=passive_deletes,
            link_model=link_model,
            sa_relationship=sa_relationship,
            sa_relationship_args=sa_relationship_args,
            sa_relationship_kwargs=sa_relationship_kwargs,
        )
        self.disable_on_redis = disable_on_redis
        self.disable_on_sql = disable_on_sql
        self.disable_on_mongo = disable_on_mongo
        self.default = default


# include sa_type, sa_column_args, sa_column_kwargs
@overload
def Field(
    default: Any = Undefined,
    *,
    default_factory: Optional["NoArgAnyCallable"] = None,
    alias: Optional[str] = None,
    title: Optional[str] = None,
    description: Optional[str] = None,
    exclude: Union[
        AbstractSet[Union[int, str]], Mapping[Union[int, str], Any], Any
    ] = None,
    include: Union[
        AbstractSet[Union[int, str]], Mapping[Union[int, str], Any], Any
    ] = None,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    allow_mutation: bool = True,
    regex: Optional[str] = None,
    discriminator: Optional[str] = None,
    repr: bool = True,
    primary_key: Union[bool, UndefinedType] = False,
    foreign_key: Any = Undefined,
    unique: Union[bool, UndefinedType] = Undefined,
    nullable: Union[bool, UndefinedType] = Undefined,
    index: Union[bool, UndefinedType] = Undefined,
    sa_type: Union[Type[Any], UndefinedType] = Undefined,
    sa_column_args: Union[Sequence[Any], UndefinedType] = Undefined,
    sa_column_kwargs: Union[Mapping[str, Any], UndefinedType] = Undefined,
    sortable: Union[bool, UndefinedType] = Undefined,
    case_sensitive: Union[bool, UndefinedType] = Undefined,
    full_text_search: Union[bool, UndefinedType] = Undefined,
    vector_options: Optional["VectorFieldOptions"] = None,
    disable_on_redis: bool = False,
    disable_on_sql: bool = False,
    disable_on_mongo: bool = False,
    schema_extra: Optional[Dict[str, Any]] = None,
) -> Any: ...


# When foreign_key is str, include ondelete
# include sa_type, sa_column_args, sa_column_kwargs
@overload
def Field(
    default: Any = Undefined,
    *,
    default_factory: Optional["NoArgAnyCallable"] = None,
    alias: Optional[str] = None,
    title: Optional[str] = None,
    description: Optional[str] = None,
    exclude: Union[
        AbstractSet[Union[int, str]], Mapping[Union[int, str], Any], Any
    ] = None,
    include: Union[
        AbstractSet[Union[int, str]], Mapping[Union[int, str], Any], Any
    ] = None,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    allow_mutation: bool = True,
    regex: Optional[str] = None,
    discriminator: Optional[str] = None,
    repr: bool = True,
    primary_key: Union[bool, UndefinedType] = False,
    foreign_key: str,
    ondelete: Union[OnDeleteType, UndefinedType] = Undefined,
    unique: Union[bool, UndefinedType] = Undefined,
    nullable: Union[bool, UndefinedType] = Undefined,
    index: Union[bool, UndefinedType] = Undefined,
    sa_type: Union[Type[Any], UndefinedType] = Undefined,
    sa_column_args: Union[Sequence[Any], UndefinedType] = Undefined,
    sa_column_kwargs: Union[Mapping[str, Any], UndefinedType] = Undefined,
    sortable: Union[bool, UndefinedType] = Undefined,
    case_sensitive: Union[bool, UndefinedType] = Undefined,
    full_text_search: Union[bool, UndefinedType] = Undefined,
    vector_options: Optional["VectorFieldOptions"] = None,
    disable_on_redis: bool = False,
    disable_on_sql: bool = False,
    disable_on_mongo: bool = False,
    schema_extra: Optional[Dict[str, Any]] = None,
) -> Any: ...


# Include sa_column, don't include
# primary_key
# foreign_key
# ondelete
# unique
# nullable
# index
# sa_type
# sa_column_args
# sa_column_kwargs
@overload
def Field(
    default: Any = Undefined,
    *,
    default_factory: Optional["NoArgAnyCallable"] = None,
    alias: Optional[str] = None,
    title: Optional[str] = None,
    description: Optional[str] = None,
    exclude: Union[
        AbstractSet[Union[int, str]], Mapping[Union[int, str], Any], Any
    ] = None,
    include: Union[
        AbstractSet[Union[int, str]], Mapping[Union[int, str], Any], Any
    ] = None,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    allow_mutation: bool = True,
    regex: Optional[str] = None,
    discriminator: Optional[str] = None,
    repr: bool = True,
    sa_column: Union["Column", UndefinedType] = Undefined,  # type: ignore
    sortable: Union[bool, UndefinedType] = Undefined,
    case_sensitive: Union[bool, UndefinedType] = Undefined,
    full_text_search: Union[bool, UndefinedType] = Undefined,
    vector_options: Optional["VectorFieldOptions"] = None,
    disable_on_redis: bool = False,
    disable_on_sql: bool = False,
    disable_on_mongo: bool = False,
    schema_extra: Optional[Dict[str, Any]] = None,
) -> Any: ...


def Field(
    default: Any = Undefined,
    *,
    default_factory: Optional["NoArgAnyCallable"] = None,
    alias: Optional[str] = None,
    title: Optional[str] = None,
    description: Optional[str] = None,
    exclude: Union[
        AbstractSet[Union[int, str]], Mapping[Union[int, str], Any], Any
    ] = None,
    include: Union[
        AbstractSet[Union[int, str]], Mapping[Union[int, str], Any], Any
    ] = None,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    allow_mutation: bool = True,
    regex: Optional[str] = None,
    discriminator: Optional[str] = None,
    repr: bool = True,
    primary_key: Union[bool, UndefinedType] = False,
    foreign_key: Any = Undefined,
    ondelete: Union[OnDeleteType, UndefinedType] = Undefined,
    unique: Union[bool, UndefinedType] = Undefined,
    nullable: Union[bool, UndefinedType] = Undefined,
    index: Union[bool, UndefinedType] = Undefined,
    sa_type: Union[Type[Any], UndefinedType] = Undefined,
    sa_column: Union["Column", UndefinedType] = Undefined,  # type: ignore
    sa_column_args: Union[Sequence[Any], UndefinedType] = Undefined,
    sa_column_kwargs: Union[Mapping[str, Any], UndefinedType] = Undefined,
    sortable: Union[bool, UndefinedType] = Undefined,
    case_sensitive: Union[bool, UndefinedType] = Undefined,
    full_text_search: Union[bool, UndefinedType] = Undefined,
    vector_options: Optional["VectorFieldOptions"] = None,
    disable_on_redis: bool = False,
    disable_on_sql: bool = False,
    disable_on_mongo: bool = False,
    schema_extra: Optional[Dict[str, Any]] = None,
) -> Any:
    current_schema_extra = schema_extra or {}
    field_info = FieldInfo(
        default,
        default_factory=default_factory,
        alias=alias,
        title=title,
        description=description,
        exclude=exclude,
        include=include,
        const=const,
        gt=gt,
        ge=ge,
        lt=lt,
        le=le,
        multiple_of=multiple_of,
        max_digits=max_digits,
        decimal_places=decimal_places,
        min_items=min_items,
        max_items=max_items,
        unique_items=unique_items,
        min_length=min_length,
        max_length=max_length,
        allow_mutation=allow_mutation,
        regex=regex,
        discriminator=discriminator,
        repr=repr,
        primary_key=primary_key,
        foreign_key=foreign_key,
        ondelete=ondelete,
        unique=unique,
        nullable=nullable,
        index=index,
        sa_type=sa_type,
        sa_column=sa_column,
        sa_column_args=sa_column_args,
        sa_column_kwargs=sa_column_kwargs,
        sortable=sortable,
        case_sensitive=case_sensitive,
        full_text_search=full_text_search,
        vector_options=vector_options,
        disable_on_redis=disable_on_redis,
        disable_on_sql=disable_on_sql,
        disable_on_mongo=disable_on_mongo,
        **current_schema_extra,
    )
    post_init_field_info(field_info)
    return field_info


@overload
def Relationship(
    *,
    back_populates: Optional[str] = None,
    cascade_delete: Optional[bool] = False,
    passive_deletes: Optional[Union[bool, Literal["all"]]] = False,
    link_model: Optional[Any] = None,
    sa_relationship_args: Optional[Sequence[Any]] = None,
    sa_relationship_kwargs: Optional[Mapping[str, Any]] = None,
    disable_on_redis: bool = False,
    disable_on_sql: bool = False,
    disable_on_mongo: bool = False,
    default: Any = Undefined,
) -> Any: ...


@overload
def Relationship(
    *,
    back_populates: Optional[str] = None,
    cascade_delete: Optional[bool] = False,
    passive_deletes: Optional[Union[bool, Literal["all"]]] = False,
    link_model: Optional[Any] = None,
    sa_relationship: Optional[RelationshipProperty[Any]] = None,
    disable_on_redis: bool = False,
    disable_on_sql: bool = False,
    disable_on_mongo: bool = False,
    default: Any = Undefined,
) -> Any: ...


def Relationship(
    *,
    back_populates: Optional[str] = None,
    cascade_delete: Optional[bool] = False,
    passive_deletes: Optional[Union[bool, Literal["all"]]] = False,
    link_model: Optional[Any] = None,
    sa_relationship: Optional[RelationshipProperty[Any]] = None,
    sa_relationship_args: Optional[Sequence[Any]] = None,
    sa_relationship_kwargs: Optional[Mapping[str, Any]] = None,
    disable_on_redis: bool = False,
    disable_on_sql: bool = False,
    disable_on_mongo: bool = False,
    default: Any = Undefined,
) -> Any:
    relationship_info = RelationshipInfo(
        back_populates=back_populates,
        cascade_delete=cascade_delete,
        passive_deletes=passive_deletes,
        link_model=link_model,
        sa_relationship=sa_relationship,
        sa_relationship_args=sa_relationship_args,
        sa_relationship_kwargs=sa_relationship_kwargs,
        disable_on_redis=disable_on_redis,
        disable_on_sql=disable_on_sql,
        disable_on_mongo=disable_on_mongo,
        default=default,
    )
    return relationship_info


def get_field_definitions(
    schema: type[ModelT],
    embedded_models: dict[str, Type] | None = None,
    relationships: dict[str, Type] | None = None,
    is_for_redis: bool = False,
    is_for_mongo: bool = False,
    is_for_sql: bool = False,
) -> dict[str, tuple[Type[Any], FieldInfo]]:
    """Retrieves the field definitions from the given schema and embedded models

    Args:
        schema: the model schema class
        embedded_models: the map of embedded models as <field_name>: <type annotation>
        relationships: the map of relationships as <field_name>: <type annotation>
        is_for_redis: whether the definitions are for redis or not
        is_for_mongo: whether the definitions are for mongo or not
        is_for_sql: whether the definitions are for sql or not

    Returns:
        dict of attributes for the model in format: <name>: (<type>, <FieldInfo>)
    """
    if embedded_models is None:
        embedded_models = {}

    if relationships is None:
        relationships = {}

    fields = {}
    for field_name, field in schema.model_fields.items():  # type: str, FieldInfo
        class_field_definition = _get_class_field_definition(field)
        if is_for_redis and class_field_definition.disable_on_redis:
            continue

        if is_for_mongo and class_field_definition.disable_on_mongo:
            continue

        if is_for_sql and class_field_definition.disable_on_sql:
            continue

        field_type = field.annotation
        field_info = field

        if field_name in embedded_models:
            field_type = embedded_models[field_name]
            field_info = copy(field)
            field_info.default = class_field_definition.default

        elif field_name in relationships:
            field_type = relationships[field_name]
            # redefine the class so that SQLModel can redo its thing
            field_info = class_field_definition

        fields[field_name] = (field_type, field_info)
    return fields


def _get_class_field_definition(field: FieldInfo) -> RelationshipInfo | FieldInfo:
    """Retrieves the relationship pr field as originally defined on the class

    SQLModel seems to change Relationship definitions to field definitions
    then hiding the real Relationship in the field as the `default` value.
    Ths function retrieves the original field definition as seen on the class definition.

    Args:
        field: the field from which to extract the relationship info

    Returns:
        the relationship info if field is relationship else it returns the field
    """
    try:
        if isinstance(field.default, RelationshipInfo):
            return field.default
    except Exception:
        pass
    return field



"""Tests for the query package"""

import pytest

from tests.conftest import RedisLibrary, SqlLibrary
from tests.utils import is_lib_installed, to_sql_text


@pytest.mark.skipif(not is_lib_installed("redis_om"), reason="Requires redis_om.")
def test_eq_redis(redis_qparser):
    """$eq checks equality in redis"""
    query = {"name": {"$eq": "Hoima, Uganda"}}
    got = redis_qparser.to_redis(RedisLibrary, query)
    assert got == ((RedisLibrary.name == "Hoima, Uganda"),)


@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
def test_eq_sql(sql_qparser):
    """$eq checks equality in sql"""
    query = {"name": {"$eq": "Hoima, Uganda"}}
    got = to_sql_text(SqlLibrary, sql_qparser.to_sql(SqlLibrary, query))
    expected = to_sql_text(SqlLibrary, ((SqlLibrary.name == "Hoima, Uganda"),))
    assert got == expected


@pytest.mark.skipif(not is_lib_installed("redis_om"), reason="Requires redis_om.")
def test_gt_redis(redis_qparser):
    """$gt checks greater than in redis"""
    query = {"name": {"$gt": "Hoima, Uganda"}}
    got = redis_qparser.to_redis(RedisLibrary, query)
    assert got == ((RedisLibrary.name > "Hoima, Uganda"),)


@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
def test_gt_sql(sql_qparser):
    """$gt checks greater than in sql"""
    query = {"name": {"$gt": "Hoima, Uganda"}}
    got = to_sql_text(SqlLibrary, sql_qparser.to_sql(SqlLibrary, query))
    expected = to_sql_text(SqlLibrary, ((SqlLibrary.name > "Hoima, Uganda"),))
    assert got == expected


@pytest.mark.skipif(not is_lib_installed("redis_om"), reason="Requires redis_om.")
def test_gte_redis(redis_qparser):
    """$gte checks greater or equal in redis"""
    query = {"name": {"$gte": "Hoima, Uganda"}}
    got = redis_qparser.to_redis(RedisLibrary, query)
    assert got == ((RedisLibrary.name >= "Hoima, Uganda"),)


@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
def test_gte_sql(sql_qparser):
    """$gte checks greater or equal in sql"""
    query = {"name": {"$gte": "Hoima, Uganda"}}
    got = to_sql_text(SqlLibrary, sql_qparser.to_sql(SqlLibrary, query))
    expected = to_sql_text(SqlLibrary, ((SqlLibrary.name >= "Hoima, Uganda"),))
    assert got == expected


@pytest.mark.skipif(not is_lib_installed("redis_om"), reason="Requires redis_om.")
def test_in_redis(redis_qparser):
    """$in checks value in list in redis"""
    query = {"name": {"$in": ["Hoima, Uganda"]}}
    got = redis_qparser.to_redis(RedisLibrary, query)
    assert got == ((RedisLibrary.name << ["Hoima, Uganda"]),)


@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
def test_in_sql(sql_qparser):
    """$in checks vlue in list than in sql"""
    query = {"name": {"$in": ["Hoima, Uganda"]}}
    got = to_sql_text(SqlLibrary, sql_qparser.to_sql(SqlLibrary, query))
    expected = to_sql_text(SqlLibrary, ((SqlLibrary.name.in_(["Hoima, Uganda"])),))
    assert got == expected


@pytest.mark.skipif(not is_lib_installed("redis_om"), reason="Requires redis_om.")
def test_lt_redis(redis_qparser):
    """$lt checks less than in redis"""
    query = {"name": {"$lt": "Hoima, Uganda"}}
    got = redis_qparser.to_redis(RedisLibrary, query)
    assert got == ((RedisLibrary.name < "Hoima, Uganda"),)


@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
def test_lt_sql(sql_qparser):
    """$lt checks less than in sql"""
    query = {"name": {"$lt": "Hoima, Uganda"}}
    got = to_sql_text(SqlLibrary, sql_qparser.to_sql(SqlLibrary, query))
    expected = to_sql_text(SqlLibrary, ((SqlLibrary.name < "Hoima, Uganda"),))
    assert got == expected


@pytest.mark.skipif(not is_lib_installed("redis_om"), reason="Requires redis_om.")
def test_lte_redis(redis_qparser):
    """$lt checks less or equal in redis"""
    query = {"name": {"$lte": "Hoima, Uganda"}}
    got = redis_qparser.to_redis(RedisLibrary, query)
    assert got == ((RedisLibrary.name <= "Hoima, Uganda"),)


@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
def test_lte_sql(sql_qparser):
    """$lt checks less or equal in sql"""
    query = {"name": {"$lte": "Hoima, Uganda"}}
    got = to_sql_text(SqlLibrary, sql_qparser.to_sql(SqlLibrary, query))
    expected = to_sql_text(SqlLibrary, ((SqlLibrary.name <= "Hoima, Uganda"),))
    assert got == expected


@pytest.mark.skipif(not is_lib_installed("redis_om"), reason="Requires redis_om.")
def test_ne_redis(redis_qparser):
    """$ne checks not equal in redis"""
    query = {"name": {"$ne": "Hoima, Uganda"}}
    got = redis_qparser.to_redis(RedisLibrary, query)
    assert got == ((RedisLibrary.name != "Hoima, Uganda"),)


@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
def test_ne_sql(sql_qparser):
    """$ne checks not equal in sql"""
    query = {"name": {"$ne": "Hoima, Uganda"}}
    got = to_sql_text(SqlLibrary, sql_qparser.to_sql(SqlLibrary, query))
    expected = to_sql_text(SqlLibrary, ((SqlLibrary.name != "Hoima, Uganda"),))
    assert got == expected


@pytest.mark.skipif(not is_lib_installed("redis_om"), reason="Requires redis_om.")
def test_nin_redis(redis_qparser):
    """$nin checks not in in redis"""
    query = {"name": {"$nin": ["Hoima, Uganda"]}}
    got = redis_qparser.to_redis(RedisLibrary, query)
    assert got == ((RedisLibrary.name >> ["Hoima, Uganda"]),)


@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
def test_nin_sql(sql_qparser):
    """$nin checks not in in sql"""
    query = {"name": {"$nin": ["Hoima, Uganda"]}}
    got = to_sql_text(SqlLibrary, sql_qparser.to_sql(SqlLibrary, query))
    expected = to_sql_text(SqlLibrary, ((SqlLibrary.name.not_in(["Hoima, Uganda"])),))
    assert got == expected


@pytest.mark.skipif(not is_lib_installed("redis_om"), reason="Requires redis_om.")
def test_not_redis(redis_qparser):
    """$not checks not in redis"""
    query = {"name": {"$not": {"$lt": "Hoima, Uganda"}}}
    got = redis_qparser.to_redis(RedisLibrary, query)
    assert got == (~(RedisLibrary.name < "Hoima, Uganda"),)


@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
def test_not_sql(sql_qparser):
    """$not checks not in sql"""
    query = {"name": {"$not": {"$lt": "Hoima, Uganda"}}}
    got = to_sql_text(SqlLibrary, sql_qparser.to_sql(SqlLibrary, query))
    expected = to_sql_text(SqlLibrary, (~(SqlLibrary.name < "Hoima, Uganda"),))
    assert got == expected


@pytest.mark.skipif(not is_lib_installed("redis_om"), reason="Requires redis_om.")
def test_regex_redis(redis_qparser):
    """$regex is not implemented for redis"""
    query = {"name": {"$regex": "^be.*", "$options": "i"}}
    with pytest.raises(
        NotImplementedError, match=r"redis text search is too inexpressive for regex.*"
    ):
        redis_qparser.to_redis(RedisLibrary, query)


@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
def test_regex_sql(sql_qparser):
    """$redis checks given item against a given regular expression"""
    query = {"name": {"$regex": "^be.*", "$options": "i"}}
    got = to_sql_text(SqlLibrary, sql_qparser.to_sql(SqlLibrary, query))
    expected = to_sql_text(
        SqlLibrary, ((SqlLibrary.name.regexp_match("(?i)^be.*", flags="i")),)
    )
    assert got == expected


@pytest.mark.skipif(not is_lib_installed("redis_om"), reason="Requires redis_om.")
def test_and_redis(redis_qparser):
    """$and checks all conditions fulfilled in redis"""
    query = {
        "$and": [
            {"name": {"$lt": "Hoima, Uganda"}},
            {"$or": [{"address": {"$eq": "Bar"}}, {"name": {"$gt": "Buliisa"}}]},
        ]
    }
    got = redis_qparser.to_redis(RedisLibrary, query)
    assert got == (
        (RedisLibrary.name < "Hoima, Uganda")
        & ((RedisLibrary.address == "Bar") | (RedisLibrary.name > "Buliisa")),
    )


@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
def test_and_sql(sql_qparser):
    """$and checks all conditions fulfilled in sql"""
    query = {
        "$and": [
            {"name": {"$lt": "Hoima, Uganda"}},
            {"$or": [{"address": {"$eq": "Bar"}}, {"name": {"$gt": "Buliisa"}}]},
        ]
    }
    got = to_sql_text(SqlLibrary, sql_qparser.to_sql(SqlLibrary, query))
    expected = to_sql_text(
        SqlLibrary,
        (
            (SqlLibrary.name < "Hoima, Uganda")
            & ((SqlLibrary.address == "Bar") | (SqlLibrary.name > "Buliisa")),
        ),
    )
    assert got == expected


@pytest.mark.skipif(not is_lib_installed("redis_om"), reason="Requires redis_om.")
def test_or_redis(redis_qparser):
    """$or checks any condition fulfilled in redis"""
    query = {
        "$or": [
            {"name": {"$lt": "Hoima, Uganda"}},
            {"$and": [{"address": {"$eq": "Bar"}}, {"name": {"$gt": "Buliisa"}}]},
        ]
    }
    got = redis_qparser.to_redis(RedisLibrary, query)
    assert got == (
        (RedisLibrary.name < "Hoima, Uganda")
        | ((RedisLibrary.address == "Bar") & (RedisLibrary.name > "Buliisa")),
    )


@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
def test_or_sql(sql_qparser):
    """$or checks any condition fulfilled in sql"""
    query = {
        "$or": [
            {"name": {"$lt": "Hoima, Uganda"}},
            {"$and": [{"address": {"$eq": "Bar"}}, {"name": {"$gt": "Buliisa"}}]},
        ]
    }
    got = to_sql_text(SqlLibrary, sql_qparser.to_sql(SqlLibrary, query))
    expected = to_sql_text(
        SqlLibrary,
        (
            (SqlLibrary.name < "Hoima, Uganda")
            | ((SqlLibrary.address == "Bar") & (SqlLibrary.name > "Buliisa")),
        ),
    )
    assert got == expected


@pytest.mark.skipif(not is_lib_installed("redis_om"), reason="Requires redis_om.")
def test_nor_redis(redis_qparser):
    """$nor checks none of the conditions is fulfilled in redis"""
    query = {
        "$nor": [
            {"name": {"$lt": "Hoima, Uganda"}},
            {"$or": [{"address": {"$eq": "Bar"}}, {"name": {"$gt": "Buliisa"}}]},
        ]
    }
    got = redis_qparser.to_redis(RedisLibrary, query)
    assert got == (
        ~(RedisLibrary.name < "Hoima, Uganda")
        & ~((RedisLibrary.address == "Bar") | (RedisLibrary.name > "Buliisa")),
    )


@pytest.mark.skipif(not is_lib_installed("sqlmodel"), reason="Requires sqlmodel.")
def test_nor_sql(sql_qparser):
    """$and checks none of the conditions is fulfilled in sql"""
    query = {
        "$nor": [
            {"name": {"$lt": "Hoima, Uganda"}},
            {"$or": [{"address": {"$eq": "Bar"}}, {"name": {"$gt": "Buliisa"}}]},
        ]
    }
    got = to_sql_text(SqlLibrary, sql_qparser.to_sql(SqlLibrary, query))
    expected = to_sql_text(
        SqlLibrary,
        (
            ~(
                (SqlLibrary.name < "Hoima, Uganda")
                | ((SqlLibrary.address == "Bar") | (SqlLibrary.name > "Buliisa"))
            ),
        ),
    )
    assert got == expected



"""Models that are saved in storage"""

from schemas import Todo, TodoList

from nqlstore import (
    EmbeddedJsonModel,
    EmbeddedMongoModel,
    JsonModel,
    MongoModel,
    SQLModel,
)

# mongo models
MongoTodo = EmbeddedMongoModel("MongoTodo", Todo)
MongoTodoList = MongoModel(
    "MongoTodoList", TodoList, embedded_models={"todos": list[MongoTodo]}
)

# redis models
RedisTodo = EmbeddedJsonModel("RedisTodo", Todo)
RedisTodoList = JsonModel(
    "RedisTodoList", TodoList, embedded_models={"todos": list[RedisTodo]}
)

# sqlite models
SqlTodoList = SQLModel(
    "SqlTodoList", TodoList, relationships={"todos": list["SqlTodo"]}
)
SqlTodo = SQLModel("SqlTodo", Todo, relationships={"parent": SqlTodoList | None})



"""Fixtures for tests"""

import os
from typing import Any

import pytest
import pytest_asyncio
from fastapi.testclient import TestClient
from models import (
    MongoTodo,
    MongoTodoList,
    RedisTodo,
    RedisTodoList,
    SqlTodo,
    SqlTodoList,
)

from nqlstore import MongoStore, RedisStore, SQLStore

TODO_LISTS: list[dict[str, Any]] = [
    {"name": "School Work"},
    {
        "name": "Home",
        "todos": [
            {"title": "Make my bed"},
            {"title": "Cook supper"},
        ],
    },
    {"name": "Boo", "todos": [{"title": "Talk endlessly till daybreak"}]},
]

_SQL_DB = "test.db"
_SQL_URL = f"sqlite+aiosqlite:///{_SQL_DB}"
_REDIS_URL = "redis://localhost:6379/0"
_MONGO_URL = "mongodb://localhost:27017"
_MONGO_DB = "testing"


@pytest.fixture
def client_with_sql():
    """The fastapi test client when SQL is enabled"""
    _reset_env()
    os.environ["SQL_URL"] = _SQL_URL

    from main import app

    yield TestClient(app)
    _reset_env()


@pytest_asyncio.fixture
async def client_with_redis():
    """The fastapi test client when redis is enabled"""
    _reset_env()
    os.environ["REDIS_URL"] = _REDIS_URL

    from main import app

    yield TestClient(app)
    _reset_env()


@pytest.fixture
def client_with_mongo():
    """The fastapi test client when mongodb is enabled"""
    _reset_env()

    os.environ["MONGO_URL"] = _MONGO_URL
    os.environ["MONGO_DB"] = _MONGO_DB

    from main import app

    yield TestClient(app)
    _reset_env()


@pytest_asyncio.fixture()
async def sql_store():
    """The sql store stored in memory"""
    store = SQLStore(uri=_SQL_URL)

    await store.register([SqlTodoList, SqlTodo])
    yield store

    # clean up
    os.remove(_SQL_DB)


@pytest_asyncio.fixture()
async def mongo_store():
    """The mongodb store. Requires a running instance of mongodb"""
    import pymongo

    mongo_store = MongoStore(uri=_MONGO_URL, database=_MONGO_DB)
    await mongo_store.register([MongoTodoList, MongoTodo])

    yield mongo_store

    # clean up
    client = pymongo.MongoClient(_MONGO_URL)  # type: ignore
    client.drop_database(_MONGO_DB)


@pytest_asyncio.fixture
async def redis_store():
    """The redis store. Requires a running instance of redis stack"""
    import redis

    store = RedisStore(_REDIS_URL)
    await store.register([RedisTodoList, RedisTodo])

    yield store

    # clean up
    client = redis.Redis("localhost", 6379, 0)
    client.flushall()


@pytest_asyncio.fixture()
async def sql_todolists(sql_store: SQLStore):
    """A list of todolists in the sql store"""
    from main import SqlTodoList

    records = await sql_store.insert(SqlTodoList, TODO_LISTS)
    yield records


@pytest_asyncio.fixture()
async def mongo_todolists(mongo_store: MongoStore):
    """A list of todolists in the mongo store"""
    from main import MongoTodoList

    records = await mongo_store.insert(MongoTodoList, TODO_LISTS)
    yield records


@pytest_asyncio.fixture()
async def redis_todolists(redis_store: RedisStore):
    """A list of todolists in the redis store"""
    from main import RedisTodoList

    records = await redis_store.insert(RedisTodoList, TODO_LISTS)
    yield records


def _reset_env():
    """Resets the environment variables available to the app"""
    os.environ["SQL_URL"] = ""
    os.environ["REDIS_URL"] = ""
    os.environ["MONGO_URL"] = ""
    os.environ["MONGO_DB"] = "testing"



import logging
from contextlib import asynccontextmanager
from typing import Annotated

from beanie import PydanticObjectId
from fastapi import Depends, FastAPI, HTTPException, Query, status
from models import MongoTodoList, RedisTodoList, SqlTodoList
from schemas import TodoList
from stores import clear_stores, get_mongo_store, get_redis_store, get_sql_store

from nqlstore import MongoStore, RedisStore, SQLStore

_SqlStoreDep = Annotated[SQLStore | None, Depends(get_sql_store)]
_RedisStoreDep = Annotated[RedisStore | None, Depends(get_redis_store)]
_MongoStoreDep = Annotated[MongoStore | None, Depends(get_mongo_store)]


@asynccontextmanager
async def lifespan(app_: FastAPI):
    clear_stores()
    yield
    clear_stores()


app = FastAPI(lifespan=lifespan)


@app.get("/todos")
async def search(
    sql: _SqlStoreDep,
    redis: _RedisStoreDep,
    mongo: _MongoStoreDep,
    q: str = Query(...),
):
    """Searches for todos by name"""

    results = []
    query = {"name": {"$regex": f".*{q}.*", "$options": "i"}}

    try:
        if sql:
            results += await sql.find(SqlTodoList, query=query)

        if redis:
            # redis's regex search is not mature so we use its full text search
            results += await redis.find(RedisTodoList, (RedisTodoList.name % f"*{q}*"))

        if mongo:
            results += await mongo.find(MongoTodoList, query=query)
    except Exception as exp:
        logging.error(exp)
        raise exp

    return [item.model_dump(mode="json") for item in results]


@app.get("/todos/{id_}")
async def get_one(
    sql: _SqlStoreDep,
    redis: _RedisStoreDep,
    mongo: _MongoStoreDep,
    id_: int | str,
):
    """Get todolist by id"""
    results = []
    query = {"id": {"$eq": id_}}

    try:
        if sql:
            results += await sql.find(SqlTodoList, query=query, limit=1)

        if redis:
            results += await redis.find(RedisTodoList, query=query, limit=1)

        if mongo:
            results += await mongo.find(
                MongoTodoList, query={"_id": {"$eq": PydanticObjectId(id_)}}, limit=1
            )
    except Exception as exp:
        logging.error(exp)
        raise exp

    try:
        return results[0].model_dump(mode="json")
    except IndexError:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)


@app.post("/todos")
async def create_one(
    sql: _SqlStoreDep,
    redis: _RedisStoreDep,
    mongo: _MongoStoreDep,
    payload: TodoList,
):
    """Create a todolist"""
    results = []
    payload_dict = payload.model_dump(exclude_unset=True)

    try:
        if sql:
            results += await sql.insert(SqlTodoList, [payload_dict])

        if redis:
            results += await redis.insert(RedisTodoList, [payload_dict])

        if mongo:
            results += await mongo.insert(MongoTodoList, [payload_dict])

        result = results[0].model_dump(mode="json")
        return result
    except Exception as exp:
        logging.error(exp)
        raise exp


@app.put("/todos/{id_}")
async def update_one(
    sql: _SqlStoreDep,
    redis: _RedisStoreDep,
    mongo: _MongoStoreDep,
    id_: int | str,
    payload: TodoList,
):
    """Update a todolist"""
    results = []
    query = {"id": {"$eq": id_}}
    updates = payload.model_dump(exclude_unset=True)

    try:
        if sql:
            results += await sql.update(SqlTodoList, query=query, updates=updates)

        if redis:
            results += await redis.update(RedisTodoList, query=query, updates=updates)

        if mongo:
            results += await mongo.update(
                MongoTodoList,
                query={"_id": {"$eq": PydanticObjectId(id_)}},
                updates=updates,
            )
    except Exception as exp:
        logging.error(exp)
        raise exp

    try:
        return results[0].model_dump(mode="json")
    except IndexError:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)


@app.delete("/todos/{id_}")
async def delete_one(
    sql: _SqlStoreDep,
    redis: _RedisStoreDep,
    mongo: _MongoStoreDep,
    id_: int | str,
):
    """Delete a todolist"""
    results = []
    query = {"id": {"$eq": id_}}

    try:
        if sql:
            results += await sql.delete(SqlTodoList, query=query)

        if redis:
            results += await redis.delete(RedisTodoList, query=query)

        if mongo:
            results += await mongo.delete(
                MongoTodoList, query={"_id": {"$eq": PydanticObjectId(id_)}}
            )
    except Exception as exp:
        logging.error(exp)
        raise exp

    try:
        return results[0].model_dump(mode="json")
    except IndexError:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)



"""Module containing the QueryPredicate to be used to filter records"""

import logging
from abc import ABC, abstractmethod
from functools import reduce
from typing import Any, Iterable, TypeVar, Union

from .._compat import Expression as _RedisExpression
from .._compat import (
    _ColumnExpressionArgument,
    _RedisField,
    _RedisModel,
    _SQLField,
    _SQLModel,
)
from .selectors import OperatorSelector, QuerySelector

_SQLFilter = _ColumnExpressionArgument[bool] | bool
_RedisFilter = Any | _RedisExpression
_T = TypeVar("_T")


class QueryPredicate(ABC):
    """This is modeled on MongoDB's query predicate for filtering records"""

    __slots__ = ("selector", "value", "parser")
    selector: str | None
    value: Any
    parser: "QueryParser"

    def __init__(
        self, selector: str | None, value: Any, parser: "QueryParser", **kwargs
    ):
        self.selector = selector
        self.value = value
        self.parser = parser

    @abstractmethod
    def to_sqlalchemy(self) -> tuple[_SQLFilter, ...]:
        """Converts this predicate to filters expected by sqlalchemy"""
        raise NotImplementedError("SQL filtering not supported")

    @abstractmethod
    def to_redis(self) -> tuple[_RedisFilter, ...]:
        """Converts this predicate to filters expected by RedisOM"""
        raise NotImplementedError("redis filtering not supported")


class OperatorPredicate(QueryPredicate, ABC):
    """The predicate for operator-value types e.g. ``{"$eq": <value>}``"""

    __slots__ = ("parent",)
    selector: str
    value: Any
    parent: Union["FieldPredicate", "NotPredicate"]

    def __init__(
        self,
        selector: str,
        value: Any,
        parent: Union["FieldPredicate", "NotPredicate"],
        **kwargs,
    ):
        super().__init__(selector, value, parser=parent.parser)
        self.parent = parent


class RootPredicate(QueryPredicate):
    """the root predicate that is the parent of all predicates"""

    __slots__ = (
        "__sql_model__",
        "__redis_model__",
    )
    value: list[QueryPredicate]
    selector: None
    __sql_model__: type[_SQLModel]
    __redis_model__: type[_RedisModel]

    def __init__(
        self,
        value: QuerySelector,
        parser: "QueryParser",
        __sql_model__: type[_SQLModel] | None = None,
        __redis_model__: type[_RedisModel] | None = None,
        **kwargs,
    ):
        self.__sql_model__ = __sql_model__
        self.__redis_model__ = __redis_model__

        kwargs.pop("selector", None)
        super().__init__(selector=None, value=None, parser=parser)
        self.value = self.parser._parse(value, parent=self)

    def to_sqlalchemy(self) -> tuple[_SQLFilter, ...]:
        parsed_values = _flatten_list([v.to_sqlalchemy() for v in self.value])
        return tuple(parsed_values)

    def to_redis(self) -> tuple[_RedisFilter, ...]:
        parsed_values = _flatten_list([v.to_redis() for v in self.value])
        return tuple(parsed_values)


class FieldPredicate(QueryPredicate):
    """Generic comparison or operator on a field

    Format::

        { <field>: <value> }
    """

    __slots__ = (
        "__sql_field__",
        "__redis_field__",
        "parent",
    )
    value: list[OperatorPredicate]
    selector: str
    __sql_field__: _SQLField
    __redis_field__: _RedisField
    parent: QueryPredicate

    def __init__(
        self,
        selector: str,
        value: OperatorSelector,
        parent: QueryPredicate,
        __sql_model__: type[_SQLModel] | None = None,
        __redis_model__: type[_RedisModel] | None = None,
        **kwargs,
    ):
        if __sql_model__:
            self.__sql_field__ = _get_sql_nested_field(
                model=__sql_model__,
                path=selector,
            )

        if __redis_model__:
            self.__redis_field__ = _get_redis_nested_field(
                model=__redis_model__,
                path=selector,
            )

        kwargs.pop("selector", None)
        super().__init__(selector=selector, value=None, parser=parent.parser)
        self.value = self.parser._parse(value, parent=self)

    def to_sqlalchemy(self) -> tuple[_SQLFilter, ...]:
        from sqlmodel import and_

        expressions = _flatten_list([v.to_sqlalchemy() for v in self.value])
        return tuple([and_(*expressions)])

    def to_redis(self) -> tuple[_RedisFilter, ...]:
        expressions = _flatten_list([v.to_redis() for v in self.value])
        return tuple([_redis_and(expressions)])


## Comparison
class EqPredicate(OperatorPredicate):
    """field is equal to value

    Format::

        { $eq: <value> }
    """

    __slots__ = ()

    def __init__(self, value: Any, parent: FieldPredicate, **kwargs):
        kwargs.pop("selector", None)
        super().__init__(selector="$eq", value=value, parent=parent, **kwargs)

    def to_sqlalchemy(self) -> tuple[_SQLFilter, ...]:
        return (self.parent.__sql_field__ == self.value,)

    def to_redis(self) -> tuple[_RedisFilter, ...]:
        return (self.parent.__redis_field__ == self.value,)


class GtPredicate(OperatorPredicate):
    """field is greater than value

    Format::

        { $gt: <value> }
    """

    __slots__ = ()

    def __init__(self, value: Any, parent: FieldPredicate, **kwargs):
        kwargs.pop("selector", None)
        super().__init__(selector="$gt", value=value, parent=parent, **kwargs)

    def to_sqlalchemy(self) -> tuple[_SQLFilter, ...]:
        return (self.parent.__sql_field__ > self.value,)

    def to_redis(self) -> tuple[_RedisFilter, ...]:
        return (self.parent.__redis_field__ > self.value,)


class GtePredicate(OperatorPredicate):
    """field is greater or equal to value

    Format::

        { $gte: <value> }
    """

    __slots__ = ()

    def __init__(self, value: Any, parent: FieldPredicate, **kwargs):
        kwargs.pop("selector", None)
        super().__init__(selector="$gte", value=value, parent=parent, **kwargs)

    def to_sqlalchemy(self) -> tuple[_SQLFilter, ...]:
        return (self.parent.__sql_field__ >= self.value,)

    def to_redis(self) -> tuple[_RedisFilter, ...]:
        return (self.parent.__redis_field__ >= self.value,)


class InPredicate(OperatorPredicate):
    """field is in list of values

    Format::

        { $in: [<value1>, <value2>, ... <valueN> ] }
    """

    __slots__ = ()

    def __init__(self, value: list[Any], parent: FieldPredicate, **kwargs):
        kwargs.pop("selector", None)
        super().__init__(selector="$in", value=value, parent=parent, **kwargs)

    def to_sqlalchemy(self) -> tuple[_SQLFilter, ...]:
        return (self.parent.__sql_field__.in_(self.value),)

    def to_redis(self) -> tuple[_RedisFilter, ...]:
        return (self.parent.__redis_field__ << self.value,)


class LtPredicate(OperatorPredicate):
    """field is less than value

    Format::

        { $lt: <value> }
    """

    __slots__ = ()

    def __init__(self, value: Any, parent: FieldPredicate, **kwargs):
        kwargs.pop("selector", None)
        super().__init__(selector="$lt", value=value, parent=parent, **kwargs)

    def to_sqlalchemy(self) -> tuple[_SQLFilter, ...]:
        return (self.parent.__sql_field__ < self.value,)

    def to_redis(self) -> tuple[_RedisFilter, ...]:
        return (self.parent.__redis_field__ < self.value,)


class LtePredicate(OperatorPredicate):
    """field is less or equal to value

    Format::

        { $lte: <value> }
    """

    __slots__ = ()

    def __init__(self, value: Any, parent: FieldPredicate, **kwargs):
        kwargs.pop("selector", None)
        super().__init__(selector="$lte", value=value, parent=parent, **kwargs)

    def to_sqlalchemy(self) -> tuple[_SQLFilter, ...]:
        return (self.parent.__sql_field__ <= self.value,)

    def to_redis(self) -> tuple[_RedisFilter, ...]:
        return (self.parent.__redis_field__ <= self.value,)


class NePredicate(OperatorPredicate):
    """field is not equal to value

    Format::

        { $ne: <value> }
    """

    __slots__ = ()

    def __init__(self, value: Any, parent: FieldPredicate, **kwargs):
        kwargs.pop("selector", None)
        super().__init__(selector="$ne", value=value, parent=parent, **kwargs)

    def to_sqlalchemy(self) -> tuple[_SQLFilter, ...]:
        return (self.parent.__sql_field__ != self.value,)

    def to_redis(self) -> tuple[_RedisFilter, ...]:
        return (self.parent.__redis_field__ != self.value,)


class RegexPredicate(OperatorPredicate):
    """field matches giv en regular expression

    Format::

        { "$regex": "pattern", "$options": "<options>" }
    """

    __slots__ = ("options",)
    options: str
    value: str

    def __init__(
        self,
        value: str,
        parent: FieldPredicate,
        raw_query: QuerySelector | OperatorSelector,
        **kwargs,
    ):
        kwargs.pop("selector", None)
        self.options = raw_query.get("$options", "")
        super().__init__(selector="$regex", value=value, parent=parent, **kwargs)

    def to_sqlalchemy(self) -> tuple[_SQLFilter, ...]:
        value = self.value
        if "sqlite" in self.parent.__sql_field__.dialect_options and self.options:
            value = f"(?{self.options}){self.value}"

        return (self.parent.__sql_field__.regexp_match(value, flags=self.options),)

    def to_redis(self) -> tuple[_RedisFilter, ...]:
        raise NotImplementedError(f"redis text search is too inexpressive for regex.")


class NinPredicate(OperatorPredicate):
    """field is not in list of values

    Format::

        { $nin: [<value1>, <value2>, ... <valueN> ] }
    """

    __slots__ = ()

    def __init__(self, value: list[Any], parent: FieldPredicate, **kwargs):
        kwargs.pop("selector", None)
        super().__init__(selector="$nin", value=value, parent=parent, **kwargs)

    def to_sqlalchemy(self) -> tuple[_SQLFilter, ...]:
        return (self.parent.__sql_field__.not_in(self.value),)

    def to_redis(self) -> tuple[_RedisFilter, ...]:
        return (self.parent.__redis_field__ >> self.value,)


class NotPredicate(QueryPredicate):
    """field is not <value> where value is a predicate e.g. { "$eq": <inner_value> }

    Format::

        { $not: <value>  }
    """

    __slots__ = ("__sql_field__", "__redis_field__", "parent")
    value: list[OperatorPredicate]
    selector: str
    __sql_field__: _SQLField
    __redis_field__: _RedisField
    parent: FieldPredicate

    def __init__(self, value: OperatorSelector, parent: FieldPredicate, **kwargs):
        if hasattr(parent, "__sql_field__"):
            self.__sql_field__ = parent.__sql_field__
        if hasattr(parent, "__redis_field__"):
            self.__redis_field__ = parent.__redis_field__
        self.parent = parent

        kwargs.pop("selector", None)
        super().__init__(selector="$not", value=None, parser=parent.parser, **kwargs)
        self.value = self.parser._parse(value, parent=self)

    def to_sqlalchemy(self) -> tuple[_SQLFilter, ...]:
        from sqlmodel import and_, not_

        expressions = _flatten_list([expr.to_sqlalchemy() for expr in self.value])
        return tuple([not_(and_(*expressions))])

    def to_redis(self) -> tuple[_RedisFilter, ...]:
        expressions = _flatten_list([expr.to_redis() for expr in self.value])
        return tuple([~(_redis_and(expressions))])


class MultiLogicalPredicate(QueryPredicate, ABC):
    """the base logical predicate like 'and', 'nor'"""

    __slots__ = ("parent",)
    value: list[list[QueryPredicate]]
    selector: str
    parent: QueryPredicate

    def __init__(
        self,
        value: list[QuerySelector],
        selector: str,
        parent: QueryPredicate,
        **kwargs,
    ):
        self.parent = parent
        super().__init__(selector=selector, value=None, parser=parent.parser, **kwargs)
        self.value = [self.parser._parse(v, parent=parent) for v in value]


class AndPredicate(MultiLogicalPredicate):
    """all expressions are true

    Format::

        { $and: [ { <expression1> }, { <expression2> } , ... , { <expressionN> } ] }
    """

    __slots__ = ()

    def __init__(self, value: list[QuerySelector], parent: FieldPredicate, **kwargs):
        kwargs.pop("selector", None)
        super().__init__(selector="$and", value=value, parent=parent, **kwargs)

    def to_sqlalchemy(self) -> tuple[_SQLFilter, ...]:
        from sqlmodel import and_

        expressions = _flatten_list(
            [expr.to_sqlalchemy() for sublist in self.value for expr in sublist]
        )
        return tuple([and_(*expressions)])

    def to_redis(self) -> tuple[_RedisFilter, ...]:
        expressions = _flatten_list(
            [expr.to_redis() for sublist in self.value for expr in sublist]
        )
        return tuple([_redis_and(expressions)])


class NorPredicate(MultiLogicalPredicate):
    """no expression is true

    Format::

        { $nor: [ { <expression1> }, { <expression2> }, ...  { <expressionN> } ] }
    """

    __slots__ = ()

    def __init__(self, value: list[QuerySelector], parent: FieldPredicate, **kwargs):
        kwargs.pop("selector", None)
        super().__init__(
            selector="$nor",
            value=value,
            parent=parent,
            **kwargs,
        )

    def to_sqlalchemy(self) -> tuple[_SQLFilter, ...]:
        from sqlmodel import not_, or_

        expressions = _flatten_list(
            [expr.to_sqlalchemy() for sublist in self.value for expr in sublist]
        )
        return tuple([not_(or_(*expressions))])

    def to_redis(self) -> tuple[_RedisFilter, ...]:
        expressions = _flatten_list(
            [expr.to_redis() for sublist in self.value for expr in sublist]
        )
        return tuple([_redis_and([~expr for expr in expressions])])


class OrPredicate(MultiLogicalPredicate):
    """any expression is true

    Format::

        { $or: [ { <expression1> }, { <expression2> }, ...  { <expressionN> } ] }
    """

    __slots__ = ()

    def __init__(self, value: list[QuerySelector], parent: FieldPredicate, **kwargs):
        kwargs.pop("selector", None)
        super().__init__(
            selector="$or",
            value=value,
            parent=parent,
            **kwargs,
        )

    def to_sqlalchemy(self) -> tuple[_SQLFilter, ...]:
        from sqlmodel import or_

        expressions = _flatten_list(
            [expr.to_sqlalchemy() for sublist in self.value for expr in sublist]
        )
        return tuple([or_(*expressions)])

    def to_redis(self) -> tuple[_RedisFilter, ...]:
        expressions = _flatten_list(
            [expr.to_redis() for sublist in self.value for expr in sublist]
        )
        return tuple([_redis_or(expressions)])


class MongoOnlyPredicate(QueryPredicate):
    """Generic predicate that handles any queries that are relevant to Mongo only"""

    __slots__ = (
        "__sql_model__",
        "__redis_model__",
        "__sql_field__",
        "__redis_field__",
        "parent",
    )
    parent: QueryPredicate

    def __init__(self, value: Any, selector: str, **kwargs):
        parent: QueryPredicate = kwargs.get("parent")
        self.parent = parent
        super().__init__(selector=selector, value=value, parser=parent.parser, **kwargs)
        for k, v in kwargs.items():
            try:
                setattr(self, k, v)
            except AttributeError:
                pass

    def to_sqlalchemy(self) -> tuple[_SQLFilter, ...]:
        logging.warning(f"{self.selector} might not work as expected in SQL databases")
        return ()

    def to_redis(self) -> tuple[_RedisFilter, ...]:
        logging.warning(f"{self.selector} might not work as expected in redis")
        return ()


class QueryParser(dict):
    """The registry of parsers of each section of the query selector

    This can be overridden to provide custom parsers for specific query selectors
    """

    _parsers: dict[str, type[QueryPredicate]] = {
        "$eq": EqPredicate,
        "$gt": GtPredicate,
        "$gte": GtePredicate,
        "$in": InPredicate,
        "$lt": LtPredicate,
        "$lte": LtePredicate,
        "$ne": NePredicate,
        "$nin": NinPredicate,
        "$not": NotPredicate,
        "$and": AndPredicate,
        "$nor": NorPredicate,
        "$or": OrPredicate,
        "$exists": MongoOnlyPredicate,
        "$type": MongoOnlyPredicate,
        "$expr": MongoOnlyPredicate,
        "$jsonSchema": MongoOnlyPredicate,
        "$mod": MongoOnlyPredicate,  # TODO: implement this for SQL and redis
        "$regex": RegexPredicate,
        "$text": MongoOnlyPredicate,  # TODO: implement this for SQL and redis, too complex
        "$where": MongoOnlyPredicate,
        "$geoIntersects": MongoOnlyPredicate,
        "$geoWithin": MongoOnlyPredicate,
        "$near": MongoOnlyPredicate,
        "$nearSphere": MongoOnlyPredicate,
        "$all": MongoOnlyPredicate,
        "$elemMatch": MongoOnlyPredicate,
        "$size": MongoOnlyPredicate,  # TODO: implement this for SQL and redis
        "$bitsAllClear": MongoOnlyPredicate,
        "$bitsAllSet": MongoOnlyPredicate,
        "$bitsAnyClear": MongoOnlyPredicate,
        "$bitsAnySet": MongoOnlyPredicate,
        "$": MongoOnlyPredicate,
        "$meta": MongoOnlyPredicate,
        "$slice": MongoOnlyPredicate,
        "$rand": MongoOnlyPredicate,
        "$natural": MongoOnlyPredicate,
        "$options": MongoOnlyPredicate,
    }

    def __init__(self, overrides: dict[str, type[QueryPredicate]] | None = None):
        """Initialize the parser registry class with any parsers overridden or new ones added

        The defaults parsers are as follows::

            {
                "$eq": :class:`~EqPredicate`,
                "$gt": :class:`~GtPredicate`,
                "$gte": :class:`~GtePredicate`,
                "$in": :class:`~InPredicate`,
                "$lt": :class:`~LtPredicate`,
                "$lte": :class:`~LtePredicate`,
                "$ne": :class:`~NePredicate`,
                "$nin": :class:`~NinPredicate`,
                "$not": :class:`~NotPredicate`,
                "$and": :class:`~AndPredicate`,
                "$nor": :class:`~NorPredicate`,
                "$or": :class:`~OrPredicate`,
                "$exists": :class:`~MongoOnlyPredicate`,
                "$type": :class:`~MongoOnlyPredicate`,
                "$expr": :class:`~MongoOnlyPredicate`,
                "$jsonSchema": :class:`~MongoOnlyPredicate`,
                "$mod": :class:`~MongoOnlyPredicate`,
                "$regex": :class:`~RegexPredicate`,
                "$text": :class:`~MongoOnlyPredicate`,
                "$where": :class:`~MongoOnlyPredicate`,
                "$geoIntersects": :class:`~MongoOnlyPredicate`,
                "$geoWithin": :class:`~MongoOnlyPredicate`,
                "$near": :class:`~MongoOnlyPredicate`,
                "$nearSphere": :class:`~MongoOnlyPredicate`,
                "$all": :class:`~MongoOnlyPredicate`,
                "$elemMatch": :class:`~MongoOnlyPredicate`,
                "$size": :class:`~MongoOnlyPredicate`,
                "$bitsAllClear": :class:`~MongoOnlyPredicate`,
                "$bitsAllSet": :class:`~MongoOnlyPredicate`,
                "$bitsAnyClear": :class:`~MongoOnlyPredicate`,
                "$bitsAnySet": :class:`~MongoOnlyPredicate`,
                "$": :class:`~MongoOnlyPredicate`,
                "$meta": :class:`~MongoOnlyPredicate`,
                "$slice": :class:`~MongoOnlyPredicate`,
                "$rand": :class:`~MongoOnlyPredicate`,
                "$natural": :class:`~MongoOnlyPredicate`,
                "$options": :class:`~MongoOnlyPredicate`,
            }

        Args:
            overrides: a dictionary with similar structure as _parsers above to override or add
                new selector parsers
        """
        if overrides:
            self._parsers = {**self._parsers, **overrides}

        super().__init__()

    def to_redis(
        self, model: type[_RedisModel], query: QuerySelector
    ) -> tuple[_SQLFilter, ...]:
        """Converts the mongodb-like NQL query to redis specific filters

        Args:
            model: the current redis model being queried
            query: the mongodb-like query

        Returns:
            the redis filters to pass to the redis finder function
        """
        root = RootPredicate(value=query, parser=self, __redis_model__=model)
        return root.to_redis()

    def to_sql(
        self, model: type[_SQLModel], query: QuerySelector
    ) -> tuple[_SQLFilter, ...]:
        """Converts the mongodb-like NQL query to SQL specific filters

        Args:
            model: the current SQL model being queried
            query: the mongodb-like query

        Returns:
            the SQL filters to pass to the SQL finder function
        """
        root = RootPredicate(value=query, parser=self, __sql_model__=model)
        return root.to_sqlalchemy()

    def _parse(
        self,
        selector: QuerySelector | OperatorSelector,
        parent: QueryPredicate,
    ) -> list[QueryPredicate]:
        """Converts the query selector to a stronger typed query predicates

        Args:
            selector: the mongo-like query filter dict
            parent: the parent of the given selector

        Returns:
            the strongly typed query predicates
        """
        # extra optional args to pass to initializer
        kwargs = dict(
            __sql_model__=getattr(parent, "__sql_model__", None),
            __redis_model__=getattr(parent, "__redis_model__", None),
            parent=parent,
            raw_query=selector,
        )

        return [
            self._get_predicate_cls(k)(selector=k, value=v, **kwargs)
            for k, v in selector.items()
        ]

    def _get_predicate_cls(self, selector: str) -> type[QueryPredicate]:
        """Gets the appropriate QueryPredicate class basing on selector

        Args:
            selector: the selector string like ``$and``, ``$not`` etc

        Returns:
            the right Predicate class
        """
        try:
            return self._parsers[selector]
        except KeyError:
            if selector.startswith("$"):
                raise NotImplementedError(
                    f"selector {selector} not supported yet. Try the native interface."
                )
            return FieldPredicate


def _flatten_list(values: Iterable[Iterable[_T]]) -> list[_T]:
    """Converts a nested iterable into a non-nested iterable

    Args:
        values: the nested iterable

    Returns:
        the flattened iterable
    """
    return [item for sublist in values for item in sublist]


def _redis_and(__filters: list[_RedisFilter]) -> _RedisFilter:
    """Merges multiple redis filters into one with 'AND' logical operator

    Args:
        __filters: the redis filters

    Returns:
        the merged AND filter
    """
    return reduce(lambda prev, curr: prev & curr, __filters)


def _redis_or(__filters: list[_RedisFilter]) -> _RedisFilter:
    """Merges multiple redis filters into one with 'OR' logical operator

    Args:
        __filters: the redis filters

    Returns:
        the merged OR filter
    """
    return reduce(lambda prev, curr: prev | curr, __filters)


def _get_sql_nested_field(model: type[_SQLModel], path: str) -> _SQLField:
    """Retrieves the SQLField at the given path, which may or may not be dotted

    Args:
        path: the path to the field where dots signify relations; example books.title
        model: the parent model

    Returns:
        the SQLField at the given path

    Raises:
        ValueError: no field '{path}' found on '{parent}'
    """
    path_segments = path.split(".")
    current_parent = model

    field = None
    for idx, part in enumerate(path_segments):
        field = getattr(current_parent, part)
        try:
            field_property = getattr(field, "property")
            property_mapper = getattr(field_property, "mapper")
            current_parent = getattr(property_mapper, "class_")
        except AttributeError as exp:
            if idx == len(path_segments) - 1:
                break
            raise exp

    if field is None:
        raise ValueError(f"no field '{path}' found on '{model}'")

    return field


def _get_redis_nested_field(model: type[_RedisModel], path: str) -> _RedisField:
    """Retrieves the RedisField at the given path, which may or may not be dotted

    Args:
        path: the path to the field where dots signify relations; example books.title
        model: the parent model

    Returns:
        the RedisField at the given path

    Raises:
        ValueError: no field '{path}' found on '{parent}'
    """
    path_segments = path.split(".")
    current_parent = model

    field = None
    for idx, part in enumerate(path_segments):
        field = getattr(current_parent, part)
        try:
            current_parent = field
        except AttributeError as exp:
            if idx == len(path_segments) - 1:
                break
            raise exp

    if field is None:
        raise ValueError(f"no field '{path}' found on '{model}'")

    return field



"""MongoDB implementation"""

import re
import sys
from typing import Any, Iterable, Mapping, TypeVar

from pydantic import BaseModel
from pydantic import Field as _Field
from pydantic.main import ModelT, create_model

from ._base import BaseStore
from ._compat import (
    AsyncIOMotorClient,
    AsyncIOMotorClientSession,
    AsyncIOMotorCollection,
    Document,
    PydanticObjectId,
    SortDirection,
    init_beanie,
)
from ._field import get_field_definitions

_T = TypeVar("_T", bound=Document)
_Filter = Mapping[str, Any] | bool
_UPDATE_OP_REGEX = re.compile(r"\$\w*")


class MongoStore(BaseStore):
    """The store that persists its data in mongo db"""

    def __init__(self, uri: str, database: str, **kwargs):
        """
        Args:
            uri: the URI of the mongodb server to connect to
            database: the name of the database
            kwargs: extra key-word args to pass to AsyncIOMotorClient
        """
        super().__init__(uri, **kwargs)
        self._client = AsyncIOMotorClient(uri, **kwargs)
        self._db = self._client[database]
        self._db_name = database

    async def register(
        self,
        models: list[type[_T]],
        allow_index_dropping: bool = False,
        recreate_views: bool = False,
        multiprocessing_mode: bool = False,
        skip_indexes: bool = False,
    ):
        """Registers the given models and runs any initialization steps

        Args:
            models: the list of Model's this store is to contain
            allow_index_dropping: if index dropping is allowed.
                Default False
            recreate_views: if views should be recreated. Default False
            multiprocessing_mode: bool - if multiprocessing mode is on
                it will patch the motor client to use process's event loop. Default False
            skip_indexes: if you want to skip working with the indexes.
                Default False
        """
        document_models = [
            cls for cls in models if not issubclass(cls, _EmbeddedMongoModel)
        ]
        await init_beanie(
            self._db,
            document_models=document_models,
            allow_index_dropping=allow_index_dropping,
            recreate_views=recreate_views,
            multiprocessing_mode=multiprocessing_mode,
            skip_indexes=skip_indexes,
        )

    async def insert(
        self,
        model: type[_T],
        items: Iterable[_T | dict],
        session: AsyncIOMotorClientSession | None = None,
        **pymongo_kwargs: Any,
    ) -> list[_T]:
        # parse them so that any default values from the model definition are added,
        # and proper validation is done
        parsed_items = [
            v if isinstance(v, model) else model.model_validate(v) for v in items
        ]
        items_as_dicts = [v.model_dump() for v in parsed_items]
        collection = self._get_collection(model)

        insert_result = await collection.insert_many(
            items_as_dicts, session=session, **pymongo_kwargs
        )
        raw_results = await collection.find(
            {"_id": {"$in": insert_result.inserted_ids}}, session=session
        ).to_list()
        return [model.model_validate(v) for v in raw_results]

    async def find(
        self,
        model: type[_T],
        query: _Filter | None = None,
        skip: int = 0,
        limit: int = 0,
        sort: None | str | list[tuple[str, SortDirection]] = None,
        session: AsyncIOMotorClientSession | None = None,
        **pymongo_kwargs: Any,
    ) -> list[_T]:
        if query is None:
            query = {}

        collection = self._get_collection(model)

        raw_results = await collection.find(
            query,
            skip=skip,
            limit=limit,
            session=session,
            sort=sort,
            **pymongo_kwargs,
        ).to_list()

        return [model.model_validate(v) for v in raw_results]

    async def update(
        self,
        model: type[_T],
        query: _Filter | None = None,
        updates: dict | None = None,
        session: AsyncIOMotorClientSession | None = None,
        upsert=False,
        **pymongo_kwargs: Any,
    ) -> list[_T]:
        if updates is None:
            updates = {}

        if query is None:
            query = {}

        mongo_updates = _to_mongo_updates(updates)

        collection = self._get_collection(model)
        query_cursor = collection.find(
            query, projection={"_id": True}, session=session, **pymongo_kwargs
        )
        ids = [v["_id"] async for v in query_cursor]

        await collection.update_many(
            query,
            update=mongo_updates,
            session=session,
            upsert=upsert,
            **pymongo_kwargs,
        )
        raw_results = collection.find(
            {"_id": {"$in": ids}}, session=session, **pymongo_kwargs
        )
        return [model.model_validate(v) async for v in raw_results]

    async def delete(
        self,
        model: type[_T],
        query: _Filter | None = None,
        session: AsyncIOMotorClientSession | None = None,
        **pymongo_kwargs: Any,
    ) -> list[_T]:
        if query is None:
            query = {}

        collection = self._get_collection(model)
        query_cursor = collection.find(query, session=session, **pymongo_kwargs)
        deleted_items = [model.model_validate(v) async for v in query_cursor]
        await collection.delete_many(query, session=session, **pymongo_kwargs)
        return deleted_items

    def _get_collection(self, model: type[_T]) -> AsyncIOMotorCollection:
        """Gets the collection for the given model

        Args:
            model: the model class whose collection is to be obtained

        Returns:
            the AsyncIOMotorCollection for the given model
        """
        collection_name = model.get_collection_name()
        return self._db[collection_name]


class _EmbeddedMongoModel(BaseModel):
    """An embedded model for mongo database

    It is never created in the database as a collection
    """

    pass


def MongoModel(
    name: str,
    schema: type[ModelT],
    /,
    embedded_models: dict[
        str,
        type[_EmbeddedMongoModel]
        | type[list[_EmbeddedMongoModel]]
        | type[ModelT]
        | type[list[ModelT]],
    ] = None,
) -> type[Document] | type[ModelT]:
    """Creates a new Mongo Model for the given schema

    A new model can be defined by::

        Model = MongoModel("Model", Schema, embedded_models = {"address": Address})

    Args:
        name: the name of the model
        schema: the schema from which the model is to be made
        embedded_models: a dict of embedded models of <field name>: annotation

    Returns:
        a Mongo model class with the given name
    """
    fields = get_field_definitions(
        schema, embedded_models=embedded_models, is_for_mongo=True
    )

    model = create_model(
        name,
        # module of the calling function
        __module__=sys._getframe(1).f_globals["__name__"],
        __doc__=schema.__doc__,
        __base__=(Document,),
        **fields,
    )

    _copy_settings(dst=model, src=schema)
    return model


def EmbeddedMongoModel(
    name: str,
    schema: type[ModelT],
    /,
    embedded_models: dict[
        str,
        type[_EmbeddedMongoModel]
        | type[list[_EmbeddedMongoModel]]
        | type[ModelT]
        | type[list[ModelT]],
    ] = None,
) -> type[_EmbeddedMongoModel] | type[ModelT]:
    """Creates a new embedded Mongo Model for the given schema

    A new model can be defined by::

        Model = EmbeddedMongoModel("Model", Schema)

    Args:
        name: the name of the model
        schema: the schema from which the model is to be made
        embedded_models: a dict of embedded models of <field name>: annotation

    Returns:
        an embedded Mongo model class with the given name
    """
    fields = get_field_definitions(
        schema, embedded_models=embedded_models, is_for_mongo=True
    )

    return create_model(
        name,
        # module of the calling function
        __module__=sys._getframe(1).f_globals["__name__"],
        __doc__=schema.__doc__,
        __base__=(_EmbeddedMongoModel,),
        **fields,
    )


class _IdOnly(BaseModel):
    """Class used for projecting only id"""

    id: PydanticObjectId = _Field(alias="_id")


def _to_mongo_updates(updates: dict[str, Any]) -> dict[str, Any]:
    """Converts the updates dict to a MongoLike update dict if it is not one already

    Args:
        updates: the update dict to convert

    Returns:
        the mongo update with update operators
    """
    for key in updates:
        if _UPDATE_OP_REGEX.match(key):
            return updates

    return {"$set": updates}


def _copy_settings(dst: type[Document], src: type[ModelT]):
    """Copies settings from source to destination

    Args:
        dst: the model to receive the settings
        src: the model or schema that contains the settings
    """
    try:
        settings_cls = getattr(src, "Settings")
        setattr(dst, "Settings", settings_cls)
    except AttributeError:
        pass



from typing import Any

import pytest
from bson import ObjectId
from conftest import TODO_LISTS
from fastapi.testclient import TestClient
from main import MongoTodoList, RedisTodoList, SqlTodoList

from nqlstore import MongoStore, RedisStore, SQLStore

_SEARCH_TERMS = ["ho", "oo", "work"]


@pytest.mark.asyncio
@pytest.mark.parametrize("todolist", TODO_LISTS)
async def test_create_sql_todolist(
    client_with_sql: TestClient,
    sql_store: SQLStore,
    todolist: dict,
):
    """POST to /todos creates a todolist in sql and returns it"""
    with client_with_sql as client:
        response = client.post("/todos", json=todolist)

        got = response.json()
        todolist_id = got["id"]
        raw_todos = todolist.get("todos", [])
        resp_todos = got["todos"]
        expected = {
            "id": todolist_id,
            "name": todolist["name"],
            "todos": [
                {
                    **raw,
                    "is_complete": "0",
                    "id": resp["id"],
                    "parent_id": todolist_id,
                }
                for raw, resp in zip(raw_todos, resp_todos)
            ],
        }

        db_query = {"id": {"$eq": todolist_id}}
        db_results = await sql_store.find(SqlTodoList, query=db_query, limit=1)
        record_in_db = db_results[0].model_dump()

        assert got == expected
        assert record_in_db == expected


@pytest.mark.asyncio
@pytest.mark.parametrize("todolist", TODO_LISTS)
async def test_create_redis_todolist(
    client_with_redis: TestClient,
    redis_store: RedisStore,
    todolist: dict,
):
    """POST to /todos creates a todolist in redis and returns it"""
    with client_with_redis as client:
        response = client.post("/todos", json=todolist)

        got = response.json()
        todolist_id = got["id"]
        raw_todos = todolist.get("todos", [])
        resp_todos = got["todos"]
        expected = {
            "id": todolist_id,
            "name": todolist["name"],
            "pk": todolist_id,
            "todos": [
                {
                    **raw,
                    "is_complete": "0",
                    "id": resp["id"],
                    "pk": resp["pk"],
                }
                for raw, resp in zip(raw_todos, resp_todos)
            ],
        }

        db_query = {"id": {"$eq": todolist_id}}
        db_results = await redis_store.find(RedisTodoList, query=db_query, limit=1)
        record_in_db = db_results[0].model_dump(mode="json")

        assert got == expected
        assert record_in_db == expected


@pytest.mark.asyncio
@pytest.mark.parametrize("todolist", TODO_LISTS)
async def test_create_mongo_todolist(
    client_with_mongo: TestClient,
    mongo_store: MongoStore,
    todolist: dict,
):
    """POST to /todos creates a todolist in redis and returns it"""
    with client_with_mongo as client:
        response = client.post("/todos", json=todolist)

        got = response.json()
        todolist_id = got["id"]
        raw_todos = todolist.get("todos", [])
        expected = {
            "id": todolist_id,
            "name": todolist["name"],
            "todos": [
                {
                    **raw,
                    "is_complete": "0",
                }
                for raw in raw_todos
            ],
        }

        db_query = {"_id": {"$eq": ObjectId(todolist_id)}}
        db_results = await mongo_store.find(MongoTodoList, query=db_query, limit=1)
        record_in_db = db_results[0].model_dump(mode="json")

        assert got == expected
        assert record_in_db == expected


@pytest.mark.asyncio
@pytest.mark.parametrize("index", range(len(TODO_LISTS)))
async def test_update_sql_todolist(
    client_with_sql: TestClient,
    sql_store: SQLStore,
    sql_todolists: list[SqlTodoList],
    index: int,
):
    """PUT to /todos/{id} updates the sql todolist of given id and returns updated version"""
    with client_with_sql as client:
        todolist = sql_todolists[index]
        id_ = todolist.id
        todos = [{**v.model_dump(), "is_complete": "1"} for v in todolist.todos]
        update = {
            "name": "some other name",
            "todos": [*todos, {"title": "another one"}, {"title": "another one again"}],
        }

        response = client.put(f"/todos/{id_}", json=update)

        got = response.json()
        expected = {
            **todolist.model_dump(mode="json"),
            **update,
            "todos": [
                {
                    **raw,
                    "id": final["id"],
                    "parent_id": final["parent_id"],
                    "is_complete": raw.get("is_complete", final["is_complete"]),
                }
                for raw, final in zip(update["todos"], got["todos"])
            ],
        }
        db_query = {"id": {"$eq": id_}}
        db_results = await sql_store.find(SqlTodoList, query=db_query, limit=1)
        record_in_db = db_results[0].model_dump(mode="json")

        assert got == expected
        assert record_in_db == expected


@pytest.mark.asyncio
@pytest.mark.parametrize("index", range(len(TODO_LISTS)))
async def test_update_redis_todolist(
    client_with_redis: TestClient,
    redis_store: RedisStore,
    redis_todolists: list[RedisTodoList],
    index: int,
):
    """PUT to /todos/{id} updates the redis todolist of given id and returns updated version"""
    with client_with_redis as client:
        todolist = redis_todolists[index]
        id_ = todolist.id
        todos = [{**v.model_dump(), "is_complete": "1"} for v in todolist.todos]
        update = {
            "name": "some other name",
            "todos": [*todos, {"title": "another one"}, {"title": "another one again"}],
        }

        response = client.put(f"/todos/{id_}", json=update)

        got = response.json()
        expected = {
            **todolist.model_dump(mode="json"),
            **update,
            "todos": [
                {
                    **raw,
                    "id": final["id"],
                    "pk": final["pk"],
                    "is_complete": raw.get("is_complete", final["is_complete"]),
                }
                for raw, final in zip(update["todos"], got["todos"])
            ],
        }
        db_query = {"id": {"$eq": id_}}
        db_results = await redis_store.find(RedisTodoList, query=db_query, limit=1)
        record_in_db = db_results[0].model_dump(mode="json")
        expected_in_db = {
            **expected,
            "todos": [
                {
                    **raw,
                    "id": final["id"],
                    "pk": final["pk"],
                }
                for raw, final in zip(expected["todos"], record_in_db["todos"])
            ],
        }

    assert got == expected
    assert record_in_db == expected_in_db


@pytest.mark.asyncio
@pytest.mark.parametrize("index", range(len(TODO_LISTS)))
async def test_update_mongo_todolist(
    client_with_mongo: TestClient,
    mongo_store: MongoStore,
    mongo_todolists: list[MongoTodoList],
    index: int,
):
    """PUT to /todos/{id} updates the mongo todolist of given id and returns updated version"""
    with client_with_mongo as client:
        todolist = mongo_todolists[index]
        id_ = todolist.id
        todos = [{**v.model_dump(), "is_complete": "1"} for v in todolist.todos]
        update = {
            "name": "some other name",
            "todos": [*todos, {"title": "another one"}, {"title": "another one again"}],
        }

        response = client.put(f"/todos/{id_}", json=update)

        got = response.json()
        expected = {
            **todolist.model_dump(mode="json"),
            **update,
            "todos": [
                {**v, "is_complete": v.get("is_complete", "0")} for v in update["todos"]
            ],
        }
        db_query = {"_id": {"$eq": id_}}
        db_results = await mongo_store.find(MongoTodoList, query=db_query, limit=1)
        record_in_db = db_results[0].model_dump(mode="json")

    assert got == expected
    assert record_in_db == expected


@pytest.mark.asyncio
@pytest.mark.parametrize("index", range(len(TODO_LISTS)))
async def test_delete_sql_todolist(
    client_with_sql: TestClient,
    sql_store: SQLStore,
    sql_todolists: list[SqlTodoList],
    index: int,
):
    """DELETE /todos/{id} deletes the sql todolist of given id and returns deleted version"""
    with client_with_sql as client:
        todolist = sql_todolists[index]
        id_ = todolist.id

        response = client.delete(f"/todos/{id_}")

        got = response.json()
        expected = todolist.model_dump(mode="json")

        db_query = {"id": {"$eq": id_}}
        db_results = await sql_store.find(SqlTodoList, query=db_query, limit=1)

        assert got == expected
        assert db_results == []


@pytest.mark.asyncio
@pytest.mark.parametrize("index", range(len(TODO_LISTS)))
async def test_delete_redis_todolist(
    client_with_redis: TestClient,
    redis_store: RedisStore,
    redis_todolists: list[RedisTodoList],
    index: int,
):
    """DELETE /todos/{id} deletes the redis todolist of given id and returns deleted version"""
    with client_with_redis as client:
        todolist = redis_todolists[index]
        id_ = todolist.id

        response = client.delete(f"/todos/{id_}")

        got = response.json()
        expected = todolist.model_dump(mode="json")

        db_query = {"id": {"$eq": id_}}
        db_results = await redis_store.find(RedisTodoList, query=db_query, limit=1)

        assert got == expected
        assert db_results == []


@pytest.mark.asyncio
@pytest.mark.parametrize("index", range(len(TODO_LISTS)))
async def test_delete_mongo_todolist(
    client_with_mongo: TestClient,
    mongo_store: MongoStore,
    mongo_todolists: list[MongoTodoList],
    index: int,
):
    """DELETE /todos/{id} deletes the mongo todolist of given id and returns deleted version"""
    with client_with_mongo as client:
        todolist = mongo_todolists[index]
        id_ = todolist.id

        response = client.delete(f"/todos/{id_}")

        got = response.json()
        expected = todolist.model_dump(mode="json")

        db_query = {"_id": {"$eq": id_}}
        db_results = await mongo_store.find(MongoTodoList, query=db_query, limit=1)

        assert got == expected
        assert db_results == []


@pytest.mark.asyncio
@pytest.mark.parametrize("index", range(len(TODO_LISTS)))
async def test_read_one_sql_todolist(
    client_with_sql: TestClient,
    sql_store: SQLStore,
    sql_todolists: list[SqlTodoList],
    index: int,
):
    """GET /todos/{id} gets the sql todolist of given id"""
    with client_with_sql as client:
        todolist = sql_todolists[index]
        id_ = todolist.id

        response = client.get(f"/todos/{id_}")

        got = response.json()
        expected = todolist.model_dump(mode="json")

        db_query = {"id": {"$eq": id_}}
        db_results = await sql_store.find(SqlTodoList, query=db_query, limit=1)
        record_in_db = db_results[0].model_dump(mode="json")

        assert got == expected
        assert record_in_db == expected


@pytest.mark.asyncio
@pytest.mark.parametrize("index", range(len(TODO_LISTS)))
async def test_read_one_redis_todolist(
    client_with_redis: TestClient,
    redis_store: RedisStore,
    redis_todolists: list[RedisTodoList],
    index: int,
):
    """GET /todos/{id} gets the redis todolist of given id"""
    with client_with_redis as client:
        todolist = redis_todolists[index]
        id_ = todolist.id

        response = client.get(f"/todos/{id_}")

        got = response.json()
        expected = todolist.model_dump(mode="json")

        db_query = {"id": {"$eq": id_}}
        db_results = await redis_store.find(RedisTodoList, query=db_query, limit=1)
        record_in_db = db_results[0].model_dump(mode="json")

        assert got == expected
        assert record_in_db == expected


@pytest.mark.asyncio
@pytest.mark.parametrize("index", range(len(TODO_LISTS)))
async def test_read_one_mongo_todolist(
    client_with_mongo: TestClient,
    mongo_store: MongoStore,
    mongo_todolists: list[MongoTodoList],
    index: int,
):
    """GET /todos/{id} gets the mongo todolist of given id"""
    with client_with_mongo as client:
        todolist = mongo_todolists[index]
        id_ = todolist.id

        response = client.get(f"/todos/{id_}")

        got = response.json()
        expected = todolist.model_dump(mode="json")

        db_query = {"_id": {"$eq": id_}}
        db_results = await mongo_store.find(MongoTodoList, query=db_query, limit=1)
        record_in_db = db_results[0].model_dump(mode="json")

        assert got == expected
        assert record_in_db == expected


@pytest.mark.asyncio
@pytest.mark.parametrize("q", _SEARCH_TERMS)
async def test_search_sql_by_name(
    client_with_sql: TestClient,
    sql_store: SQLStore,
    sql_todolists: list[SqlTodoList],
    q: str,
):
    """GET /todos?q={} gets all sql todolists with name containing search item"""
    with client_with_sql as client:
        response = client.get(f"/todos?q={q}")

        got = response.json()
        expected = [
            v.model_dump(mode="json") for v in sql_todolists if q in v.name.lower()
        ]

        assert got == expected


@pytest.mark.asyncio
@pytest.mark.parametrize("q", _SEARCH_TERMS)
async def test_search_redis_by_name(
    client_with_redis: TestClient,
    redis_store: RedisStore,
    redis_todolists: list[RedisTodoList],
    q: str,
):
    """GET /todos?q={} gets all redis todolists with name containing search item"""
    with client_with_redis as client:
        response = client.get(f"/todos?q={q}")

        got = response.json()
        expected = [
            v.model_dump(mode="json") for v in redis_todolists if q in v.name.lower()
        ]

        assert got == expected


@pytest.mark.asyncio
@pytest.mark.parametrize("q", _SEARCH_TERMS)
async def test_search_mongo_by_name(
    client_with_mongo: TestClient,
    mongo_store: MongoStore,
    mongo_todolists: list[MongoTodoList],
    q: str,
):
    """GET /todos?q={} gets all mongo todolists with name containing search item"""
    with client_with_mongo as client:
        response = client.get(f"/todos?q={q}")

        got = response.json()
        expected = [
            v.model_dump(mode="json") for v in mongo_todolists if q in v.name.lower()
        ]

        assert got == expected


def _get_id(item: Any) -> Any:
    """Gets the id of the given record

    Args:
        item: the record whose id is to be obtained

    Returns:
        the id of the record
    """
    try:
        return item.id
    except AttributeError:
        return item.pk
